# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023-2025 Oxford Quantum Circuits Ltd
from dataclasses import dataclass
from enum import Enum
from typing import Optional

from qat.purr.compiler.devices import Qubit, Resonator
from qat.purr.compiler.waveforms import AbstractWaveform
from qat.purr.integrations.qasm import (
    extern_port_name,
    get_frame_mappings,
    get_port_mappings,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class FeatureMetadata:
    """Exposed metadata for various languages, integrations and features."""

    def to_json_dict(self):
        """
        Turn object into a JSON-amenable dictionary to return from a web service. All
        returned dictionaries should be able to be merged together without additional
        effort, so make sure a root node with the name of the feature is available.
        """
        pass


class Scale(Enum):
    """
    SI units of frequency
    """

    NANO = "ns"
    MICRO = "us"
    MILLI = "ms"
    DEFAULT = ""
    KILO = "k"
    MEGA = "M"
    GIGA = "G"
    TERA = "T"


class Unit(Enum):
    """
    Physical SI units.
    """

    TIME = "s"
    FREQUENCY = "Hz"


# TODO: You don't want something lke this, just have two different objects for
#   time/frequency.
@dataclass(frozen=True)
class Quantity:
    amount: float
    unit: Unit
    scale: Scale

    def __str__(self):
        return f"{self.amount} {self.scale.value}{self.unit.value}"


@dataclass(frozen=True)
class Constraints:
    pulse_control_contraints: str = None
    max_scale: float = 1
    max_waveform_amplitude: float = 1
    awg_frequency_bandwidth: str = Quantity(1, Unit.FREQUENCY, Scale.DEFAULT)
    max_pulse_duration: str = Quantity(120, Unit.TIME, Scale.MICRO)
    min_pulse_duration: str = Quantity(8, Unit.TIME, Scale.NANO)
    min_pulse_time: str = Quantity(8, Unit.TIME, Scale.NANO)


class OpenPulseFeatures(FeatureMetadata):
    def __init__(self):
        self.ports = dict()
        self.frames = dict()
        self.waveforms = dict()
        self.constraints: Optional[Constraints] = None

    def for_hardware(self, hardware):
        def _find_qubit(qubit) -> Qubit:
            if isinstance(qubit, Resonator):
                for qb in hardware.qubits:
                    if qb.measure_device == qubit:
                        qubit = qb

            if isinstance(qubit, Resonator):
                raise ValueError(f"{str(qubit)} was unable to be matched to a qubit.")

            return qubit

        for frame_name, channel_view in get_frame_mappings(hardware).items():
            frame = channel_view.pulse_channel
            qubit = _find_qubit(hardware.get_devices_from_pulse_channel(frame)[0])
            qubits = [qubit.id]
            qubits.extend(
                [
                    qubit.id
                    for qubit in channel_view.auxiliary_devices
                    if isinstance(qubit, Qubit)
                ]
            )
            self.frames[frame_name] = dict(
                qubits=qubits,
                port_id=extern_port_name(frame.physical_channel),
                frequency=frame.frequency,
                bandwidth_centre=frame.baseband_frequency,
                phase=0.0,
            )

        for port_name, port in get_port_mappings(hardware).items():
            self.ports[port_name] = dict(
                direction="two-way" if port.acquire_allowed else "one-way",
                type="port_type_1",
                associated_qubits=[
                    _find_qubit(qb).index
                    for qb in hardware.quantum_devices.values()
                    if qb.physical_channel == port
                ],
            )

        self.waveforms = {
            key: vars(value.waveform_definition)
            for key, value in AbstractWaveform.actual_waveforms.items()
        }

        self.constraints = Constraints()

    def to_json_dict(self):
        return {
            "open_pulse": {
                "version": "0",
                "ports": self.ports,
                "frames": self.frames,
                "waveforms": self.waveforms,
                "constraints": self.constraints,
            }
        }



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from pathlib import Path

from compiler_config.config import CompilerConfig

from qat.core.config.configure import get_qatconfig, override_config
from qat.core.config.session import QatSessionConfig
from qat.core.metrics_base import MetricsManager
from qat.core.pipeline import EngineSet, HardwareLoaders, PipelineManager
from qat.executables import Executable
from qat.pipelines import get_default_pipelines
from qat.pipelines.base import AbstractPipeline
from qat.pipelines.pipeline import Pipeline
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.qatconfig import QatConfig


class QAT:
    def __init__(self, qatconfig: QatSessionConfig | QatConfig | str | Path | None = None):
        if isinstance(qatconfig, (str, Path)):
            qatconfig = QatSessionConfig.from_yaml(qatconfig)

        if type(qatconfig) is QatConfig:
            qatconfig = QatSessionConfig(**qatconfig.model_dump(exclude_defaults=True))

        self.config = qatconfig or QatSessionConfig(
            **get_qatconfig().model_dump(exclude_defaults=True)
        )

        self._available_hardware = HardwareLoaders.from_descriptions(self.config.HARDWARE)
        self._engines = EngineSet.from_descriptions(
            self.config.ENGINES, self._available_hardware
        )

        if (
            self.config.PIPELINES is None
            and self.config.COMPILE is None
            and self.config.EXECUTE is None
        ):
            full_pipelines, compile_pipelines, execute_pipelines = get_default_pipelines()
        else:
            full_pipelines = self.config.PIPELINES
            compile_pipelines = self.config.COMPILE
            execute_pipelines = self.config.EXECUTE

        self.pipelines = PipelineManager.from_descriptions(
            compile_pipelines=compile_pipelines,
            execute_pipelines=execute_pipelines,
            full_pipelines=full_pipelines,
            available_hardware=self._available_hardware,
            available_engines=self._engines,
        )

    def compile(
        self,
        program,
        compiler_config: CompilerConfig | None = None,
        pipeline: Pipeline | str = "default",
        to_json: bool = False,
        **kwargs,
    ) -> tuple[InstructionBuilder | Executable | str, MetricsManager]:
        """Compiles a source program into an executable using the specified pipeline.

        :param program: The source program to compile.
        :param compiler_config: Configuration options for the compiler, such as optimization
            and results formatting.
        :param pipeline: The pipeline to use for compilation. Defaults to "default".
        :param to_json: If True, the output package will be serialized to JSON format.
        :return: A tuple containing the executable of the compiled program for the target
            device and the metrics manager containing metrics collected during compilation.
        """

        with override_config(self.config):
            P = self.pipelines.get_compile_pipeline(pipeline)
            package, metrics_manager = P.compile(
                program, compiler_config=compiler_config, **kwargs
            )
            package = package.serialize() if to_json else package
            return package, metrics_manager

    def execute(
        self,
        package: InstructionBuilder | Executable | str,
        compiler_config: CompilerConfig | None = None,
        pipeline: Pipeline | str = "default",
        **kwargs,
    ) -> tuple[dict, MetricsManager]:
        """Executes a compiled package on the specified pipeline.

        :param package: The compiled package to execute, which can be provided as a JSON
            blob.
        :param compiler_config: Configuration options for the compiler, such as optimization
            and results formatting.
        :param pipeline: The pipeline to use for execution. Defaults to "default".
        :return: A tuple containing the results of the execution and the metrics manager
        """

        with override_config(self.config):
            P = self.pipelines.get_execute_pipeline(pipeline)
            if isinstance(package, str):
                # If the package is a string, deserialize it
                if "py/object" in package:
                    package = InstructionBuilder.deserialize(package)
                else:
                    package = Executable.deserialize(package)
            return P.execute(package, compiler_config=compiler_config, **kwargs)

    def run(
        self,
        program,
        compiler_config: CompilerConfig | None = None,
        pipeline: AbstractPipeline | str = "default",
        compile_pipeline: AbstractPipeline | str | None = None,
        execute_pipeline: AbstractPipeline | str | None = None,
        **kwargs,
    ) -> tuple[dict, MetricsManager]:
        """Compiles and executes a source program using the specified pipeline.

        The compilation and execution pipeline can be specified separately using
        `compile_pipeline` and `execute_pipeline` which take precedence . Alternatively,
        a unified pipeline can be provided using the `pipeline` parameter. If neither are
        provided, the default pipeline will be used for both compilation and execution.

        :param program: The source program to compile and execute.
        :param compiler_config: Configuration options for the compiler, such as optimization
            and results formatting.
        :param pipeline: The pipeline to use for compilation and execution. Defaults to
            "default".
        :param compile_pipeline: The pipeline to use for compilation. If provided, it will
            override the `pipeline` parameter for compilation.
        :param execute_pipeline: The pipeline to use for execution. If provided, it will
            override the `pipeline` parameter for execution.
        :return: A tuple containing the results of the execution and the metrics manager
        """

        compile_pipeline = compile_pipeline if compile_pipeline is not None else pipeline
        execute_pipeline = execute_pipeline if execute_pipeline is not None else pipeline

        pkg, compile_metrics = self.compile(
            program,
            compiler_config=compiler_config,
            pipeline=compile_pipeline,
            **kwargs,
        )
        result, execute_metrics = self.execute(
            pkg,
            compiler_config=compiler_config,
            pipeline=execute_pipeline,
            **kwargs,
        )
        return result, execute_metrics.merge(compile_metrics)

    def reload_all_models(self):
        """Reloads all hardware models and updates the pipelines."""
        self._available_hardware.reload_all_models()
        self._engines.reload_all_models()
        self.pipelines.reload_all_models()

    @property
    def models_up_to_date(self) -> bool:
        """Checks if all hardware models are up-to-date.

        :return: True if all models are up-to-date, False otherwise.
        """
        return self._available_hardware.models_up_to_date



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import random
from functools import cached_property
from pathlib import Path

import piny
from pydantic import (
    NonNegativeFloat,
    NonNegativeInt,
    PositiveFloat,
    PositiveInt,
    model_validator,
)

from qat.utils.piny import VeryStrictMatcher
from qat.utils.pydantic import NoExtraFieldsFrozenModel


class DeviceDescription(NoExtraFieldsFrozenModel):
    """
    Device-related target description.

    :param sample_time: The rate at which the pulse is sampled (in s).
    :param samples_per_clock_cycle: The number of samples per clock cycle.
    :param instruction_memory_size: The max. allowed number of instructions.
    :param waveform_memory_size: The max. memory that can be used for waveforms, in clock cycles.
    :param pulse_duration_min: The minimal pulse duration for all pulse channels.
    :param pulse_duration_max: The maximal pulse duration for all pulse channels.
    :param pulse_channel_lo_freq_min: The minimal LO frequency for a pulse channel.
    :param pulse_channel_lo_freq_max: The maximal LO frequency for a pulse channel.
    :param pulse_channel_if_freq_min: The minimal intermediate frequency for a pulse channel.
    :param pulse_channel_if_freq_max: The maximal intermediate frequency for a pulse channel.
    """

    sample_time: PositiveFloat
    samples_per_clock_cycle: PositiveInt
    instruction_memory_size: PositiveInt
    waveform_memory_size: PositiveInt
    pulse_duration_min: PositiveFloat
    pulse_duration_max: PositiveFloat
    pulse_channel_lo_freq_min: NonNegativeInt
    pulse_channel_lo_freq_max: NonNegativeInt
    pulse_channel_if_freq_min: NonNegativeInt
    pulse_channel_if_freq_max: NonNegativeInt

    @model_validator(mode="after")
    def validate_durations(self):
        if self.pulse_duration_min > self.pulse_duration_max:
            raise ValueError(
                "Min. pulse duration cannot be larger than max. pulse duration."
            )
        return self

    @classmethod
    def random(cls, seed=42):
        return cls(
            **{
                "sample_time": random.Random(seed).choice([1e-09, 2e-09]),
                "samples_per_clock_cycle": random.Random(seed).randint(5, 10),
                "instruction_memory_size": random.Random(seed).randint(200, 8000),
                "waveform_memory_size": random.Random(seed).randint(1000, 2000),
                "pulse_duration_min": random.Random(seed).uniform(1e-09, 1e-07),
                "pulse_duration_max": random.Random(seed).uniform(1e-06, 1e-03),
                "pulse_channel_lo_freq_min": random.Random(seed).randint(1_000, 10_000),
                "pulse_channel_lo_freq_max": random.Random(seed).randint(
                    1_000_000, 100_000_000
                ),
                "pulse_channel_if_freq_min": random.Random(seed).randint(1_000, 10_000),
                "pulse_channel_if_freq_max": random.Random(seed).randint(
                    1_000_000, 100_000_000
                ),
            }
        )

    @cached_property
    def clock_cycle(self):
        return self.samples_per_clock_cycle * self.sample_time


class QubitDescription(DeviceDescription):
    """
    Qubit-related target description.

    :param passive_reset_time: The amount of time after each shot where the qubit is idle.
    """

    passive_reset_time: NonNegativeFloat

    @classmethod
    def random(cls, seed=42):
        device_descr = DeviceDescription.random(seed).model_dump()
        device_descr.update(
            {"passive_reset_time": random.Random(seed).uniform(1e-06, 1e-04)}
        )
        return cls(**device_descr)


class ResonatorDescription(DeviceDescription):
    """
    Resonator-related target description.
    """

    ...


class AbstractTargetData(NoExtraFieldsFrozenModel):
    """
    Data related to a general target machine.

    :param max_shots: The maximum amount of shots possible on this target.
    :param default_shots: The default amount of shots on this target if none specified through the instructions.
    """

    max_shots: PositiveInt
    default_shots: PositiveInt = 1

    def __getattribute__(self, name: str):
        if name in ["QUBIT_DATA", "RESONATOR_DATA"]:
            try:
                return super().__getattribute__(name)
            except AttributeError:
                raise AttributeError(
                    f"Tried to get '{name}' from {self.__class__.__name__}, which does not exist. Please use a child class of `AbstractTargetData` that has '{name}'."
                )
        return super().__getattribute__(name)

    @classmethod
    def from_yaml(cls, path: str | Path):
        blob = piny.YamlLoader(path=str(path), matcher=VeryStrictMatcher).load()
        return cls(**blob)


class TargetData(AbstractTargetData):
    """
    Data related to a general target machine.

    :param max_shots: The maximum amount of shots possible on this target.
    :param default_shots: The default amount of shots on this target if none specified through the instructions.
    :param QUBIT_DATA: Qubit-related target description.
    :param RESONATOR_DATA: Resonator-related target description.
    """

    QUBIT_DATA: QubitDescription
    RESONATOR_DATA: ResonatorDescription

    @cached_property
    def clock_cycle(self):
        return max(self.QUBIT_DATA.clock_cycle, self.RESONATOR_DATA.clock_cycle)

    @cached_property
    def instruction_memory_size(self):
        return max(
            self.QUBIT_DATA.instruction_memory_size,
            self.RESONATOR_DATA.instruction_memory_size,
        )

    @model_validator(mode="after")
    def validate_clock_cycles(self):
        if self.QUBIT_DATA.clock_cycle != self.RESONATOR_DATA.clock_cycle:
            raise ValueError(
                "Different clock cycles for qubit and resonator are currently not supported."
            )
        return self

    @model_validator(mode="after")
    def validate_instruction_memory_size(self):
        if (
            self.QUBIT_DATA.instruction_memory_size
            != self.RESONATOR_DATA.instruction_memory_size
        ):
            raise ValueError(
                "Different instruction memory sizes for qubit and resonator are currently not supported."
            )
        return self

    @classmethod
    def random(cls, seed: int = 42) -> "TargetData":
        return cls(
            max_shots=random.Random(seed).randint(200, 1000),
            default_shots=random.Random(seed).randint(1, 100),
            QUBIT_DATA=QubitDescription.random(seed),
            RESONATOR_DATA=ResonatorDescription.random(seed),
        )

    @classmethod
    def create_with(
        cls,
        max_shots: int = 10_000,
        default_shots: int = 1_000,
        passive_reset_time: float = 1e-03,
    ) -> "TargetData":
        """Returns a default TargetData instance, allowing configuration over user-level
        parameters.

        :param max_shots: The maximum amount of shots possible on this target.
        :param default_shots: The default amount of shots on this target if none specified
            through the instructions or compiler_config.
        :param passive_reset_time: The amount of time to wait after each shot to allow the
            qubits to passively reset.
        """
        return TargetData(
            max_shots=max_shots,
            default_shots=default_shots,
            QUBIT_DATA=QubitDescription(
                passive_reset_time=passive_reset_time,
                sample_time=1e-09,
                samples_per_clock_cycle=1,
                instruction_memory_size=50_000,
                waveform_memory_size=1_500,
                pulse_duration_min=64e-09,
                pulse_duration_max=1e-03,
                pulse_channel_lo_freq_min=1e06,
                pulse_channel_lo_freq_max=1e10,
                pulse_channel_if_freq_min=0,
                pulse_channel_if_freq_max=1e10,
            ),
            RESONATOR_DATA=ResonatorDescription(
                sample_time=1e-09,
                samples_per_clock_cycle=1,
                instruction_memory_size=50_000,
                waveform_memory_size=1_500,
                pulse_duration_min=64e-09,
                pulse_duration_max=1e-03,
                pulse_channel_lo_freq_min=1e06,
                pulse_channel_lo_freq_max=1e10,
                pulse_channel_if_freq_min=0,
                pulse_channel_if_freq_max=1e10,
            ),
        )

    @classmethod
    def default(cls) -> "TargetData":
        """Returns a default TargetData instance."""
        return cls.create_with()


def DefaultTargetData() -> TargetData:
    """Returns a default TargetData instance."""
    return TargetData.default()


def CustomTargetData(
    max_shots: int = 10_000, default_shots: int = 1_000, passive_reset_time: float = 1e-03
) -> TargetData:
    """Returns a default TargetData instance, allowing configuration over user-level
    parameters.

    :param max_shots: The maximum amount of shots possible on this target.
    :param default_shots: The default amount of shots on this target if none specified
        through the instructions or compiler_config.
    :param passive_reset_time: The amount of time to wait after each shot to allow the
        qubits to passively reset.
    """
    return TargetData.create_with(
        max_shots=max_shots,
        default_shots=default_shots,
        passive_reset_time=passive_reset_time,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.backend.fallthrough import FallthroughBackend
from qat.frontend import AutoFrontend
from qat.middleend import CustomMiddleend
from qat.model.target_data import TargetData
from qat.pipelines.pipeline import Pipeline
from qat.pipelines.updateable import UpdateablePipeline
from qat.purr.backends.live import LiveHardwareModel
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.utils.logger import get_default_logger
from qat.runtime.legacy import LegacyRuntime

from .compile import middleend_pipeline
from .execute import LegacyPipelineConfig, results_pipeline

log = get_default_logger()


class LegacyPipeline(UpdateablePipeline):
    """A pipeline that compiles programs using the legacy backend and executes them using
    the :class:`LegacyRuntime`.

    The piepline uses the engine provided by the legacy model, and cannot be provided to
    the factory.
    """

    @staticmethod
    def _build_pipeline(
        config: LegacyPipelineConfig,
        model: QuantumHardwareModel,
        target_data: TargetData | None = None,
        engine: None = None,
    ) -> Pipeline:
        if engine is not None:
            log.warning(
                "The engine for the LegacyPipeline is expected to be provided by the "
                "model, and the provided engine will be ignored."
            )

        if isinstance(model, LiveHardwareModel):
            # Let the Runtime handle startup based on ConnectionMode
            engine = model.create_engine(startup_engine=False)
        else:
            engine = model.create_engine()

        return Pipeline(
            name=config.name,
            model=model,
            target_data=target_data if target_data is not None else TargetData.default(),
            frontend=AutoFrontend.default_for_legacy(model),
            middleend=CustomMiddleend(
                model,
                pipeline=middleend_pipeline(model=model, target_data=target_data),
            ),
            backend=FallthroughBackend(model),
            runtime=LegacyRuntime(
                engine=engine,
                results_pipeline=results_pipeline(model),
                connection_mode=config.connection_mode,
            ),
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from collections import defaultdict
from dataclasses import dataclass, field

from qat.core.metrics_base import MetricsManager, MetricsType
from qat.core.pass_base import AnalysisPass, ResultManager
from qat.core.result_base import ResultInfoMixin
from qat.ir.instruction_builder import QuantumInstructionBuilder
from qat.ir.measure import Acquire
from qat.ir.waveforms import Pulse
from qat.model.device import PulseChannel, Qubit
from qat.model.hardware_model import PhysicalHardwareModel
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


@dataclass
class ActivePulseChannelResults(ResultInfoMixin):
    """Stores the active pulse channels in a task, which is defined by pulse channels that
    are acted on by a pulse or acquisition.

    Results are stored as a map between pulse channels and the qubit they belong to. Various
    helper properties and methods can be used to fetch the complete lists of active pulse
    channels and qubits.
    """

    pulse_channel_to_qubit_map: dict[str, Qubit] = field(default_factory=dict)
    qubit_to_pulse_channel_map: dict[Qubit, list[PulseChannel]] = field(
        default_factory=lambda: defaultdict(list)
    )

    @property
    def targets(self) -> set[str]:
        """Returns a dictionary of all pulse channels with their full id as a key."""
        return set(self.pulse_channel_to_qubit_map.keys())

    @property
    def qubits(self) -> set[Qubit]:
        """Returns a list of all active qubits."""
        return set(self.qubit_to_pulse_channel_map.keys())

    def add_target(self, pulse_channel: PulseChannel, qubit: Qubit):
        """Adds a pulse channel and its associated qubit to the results."""
        self.pulse_channel_to_qubit_map[pulse_channel.uuid] = qubit
        self.qubit_to_pulse_channel_map[qubit].append(pulse_channel)

    def from_qubit(self, qubit: Qubit) -> list[PulseChannel]:
        """Returns the list of pulse channels that belong to a qubit."""
        return self.qubit_to_pulse_channel_map.get(qubit)


class ActivePulseChannelAnalysis(AnalysisPass):
    """Determines the set of pulse channels which are targeted by quantum instructions.

    A pulse channel that has a pulse played at any time, or an acquisition is defined to be
    active. This pass is used to determine which pulse channels are required in compilation,
    and is used in subsequent passes to easily extract pulse channel properties, and is
    useful for not performing extra analysis on rogue channels picked up by
    :class:`Synchronize` instructions.

    Also records the physical qubit indices used in the circuit as a metric.
    """

    def __init__(self, model: PhysicalHardwareModel):
        self.model = model

    def run(
        self,
        ir: QuantumInstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        *args,
        **kwargs,
    ) -> QuantumInstructionBuilder:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to store the analysis results.
        """

        targets: set[str] = set()
        for inst in ir.instructions:
            if isinstance(inst, (Acquire, Pulse)):
                targets.add(inst.target)

        result = ActivePulseChannelResults()
        for target in targets:
            pulse_channel = ir.get_pulse_channel(target)
            device = self.model.qubit_for_physical_channel_id(
                pulse_channel.physical_channel_id
            )

            if device is None:
                raise ValueError(
                    f"Pulse channel with id {target} cannot be mapped to a device in the "
                    "hardware model."
                )
            result.add_target(pulse_channel, device)

        phys_q_indices = sorted(
            [self.model.index_of_qubit(qubit) for qubit in result.qubits]
        )
        log.info(f"Physical qubits used in this circuit: {phys_q_indices}")
        met_mgr.record_metric(MetricsType.PhysicalQubitIndices, phys_q_indices)

        res_mgr.add(result)
        return ir


PydActivePulseChannelAnalysis = ActivePulseChannelAnalysis



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.frontend.parsers.qasm.qasm2 import CloudQasmParser
from qat.frontend.parsers.qasm.qasm3 import Qasm3Parser


def get_qasm_parser(qasm_str: str):
    """Gets the appropriate QASM parser for the passed-in QASM string."""
    parsers = [CloudQasmParser(), Qasm3Parser()]
    attempts = []
    for parser in parsers:
        parse_attempt = parser.can_parse(qasm_str)
        if parse_attempt:
            return parser
        attempts.append((parser.parser_language().name, parse_attempt.errors))

    raise ValueError(
        "No valid parser could be found. Attempted: "
        f"{', '.join([f'{a} with error {b}' for a, b in attempts])}"
    )


def qasm_from_file(file_path):
    """Get QASM from a file."""
    with open(file_path) as ifile:
        return ifile.read()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from dataclasses import dataclass, field

from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


@dataclass
class ConnectionConfig:
    bulk_value: list[str] = field(default_factory=list)
    out0: str | None = None  # Enum: {‘I’, ‘Q’, ‘off’}
    out1: str | None = None  # Enum: {‘I’, ‘Q’, ‘off’}
    out2: str | None = None  # Enum: {‘I’, ‘Q’, ‘off’}
    out3: str | None = None  # Enum: {‘I’, ‘Q’, ‘off’}
    acq: str | None = None  # Enum: {False, ‘in0’, True, ‘off’}
    acq_I: str | None = None  # Enum: {‘in0’, ‘in1’, ‘off’}
    acq_Q: str | None = None  # Enum: {‘in0’, ‘in1’, ‘off’}


@dataclass
class NcoConfig:
    freq: float | None = None
    phase_offs: float | None = None
    prop_delay_comp: int | None = None
    prop_delay_comp_en: bool | None = None


@dataclass
class AwgConfig:
    cont_mode_en_path0: bool | None = None
    cont_mode_en_path1: bool | None = None
    cont_mode_waveform_idx_path0: int | None = None
    cont_mode_waveform_idx_path1: int | None = None

    upsample_rate_path0: int | None = None
    upsample_rate_path1: int | None = None

    gain_path0: float | None = None
    gain_path1: float | None = None

    offset_path0: float | None = None
    offset_path1: float | None = None

    mod_en: bool | None = None


@dataclass
class MixerConfig:
    phase_offset: float | None = None
    gain_ratio: float | None = None


@dataclass
class SquareWeightAcq:
    integration_length: int | None = None


@dataclass
class ThresholdedAcqConfig:
    rotation: float | None = None
    threshold: float | None = None
    marker_en: bool | None = None
    marker_address: int | None = None
    marker_invert: bool | None = None
    trigger_en: bool | None = None
    trigger_address: int | None = None
    trigger_invert: bool | None = None


@dataclass
class TtlAcqConfig:
    auto_bin_incr_en: bool | None = None
    threshold: float | None = None
    input_select: int | None = None


@dataclass
class SequencerConfig:
    sync_en: bool | None = None
    marker_ovr_en: bool | None = None
    marker_ovr_value: int | None = None

    trigger_count_thresholds: dict[int, float] = field(
        default_factory=dict
    )  # key in [0,15]
    trigger_threshold_inverts: dict[int, bool] = field(
        default_factory=dict
    )  # key in [0,15]

    connection: ConnectionConfig = field(default_factory=lambda: ConnectionConfig())
    nco: NcoConfig = field(default_factory=lambda: NcoConfig())
    awg: AwgConfig = field(default_factory=lambda: AwgConfig())
    mixer: MixerConfig = field(default_factory=lambda: MixerConfig())

    demod_en_acq: bool | None = None
    square_weight_acq: SquareWeightAcq = field(default_factory=lambda: SquareWeightAcq())
    thresholded_acq: ThresholdedAcqConfig = field(
        default_factory=lambda: ThresholdedAcqConfig()
    )
    ttl_acq: TtlAcqConfig = field(default_factory=lambda: TtlAcqConfig())


@dataclass
class OffsetConfig:
    out0: float | None = None
    out1: float | None = None
    out2: float | None = None
    out3: float | None = None

    in1: float | None = None
    in0: float | None = None

    out0_path0: float | None = None
    out0_path1: float | None = None
    out1_path0: float | None = None
    out1_path1: float | None = None

    in0_path0: float | None = None
    in0_path1: float | None = None


@dataclass
class LoConfig:
    out0_en: bool | None = None
    out0_freq: float | None = None
    out1_en: bool | None = None
    out1_freq: float | None = None

    out0_in0_en: bool | None = None
    out0_in0_freq: float | None = None


@dataclass
class AttConfig:
    out0: int | None = None
    out1: int | None = None
    in0: int | None = None


@dataclass
class GainConfig:
    in0: int | None = None
    in1: int | None = None


@dataclass
class ScopeAcqConfig:
    sequencer_select: int | None = None
    trigger_mode_path0: str | None = None  # Enum: {‘sequencer’, ‘level’}
    trigger_mode_path1: str | None = None  # Enum: {‘sequencer’, ‘level’}
    trigger_level_path0: float | None = None
    trigger_level_path1: float | None = None
    avg_mode_en_path0: bool | None = None
    avg_mode_en_path1: bool | None = None


@dataclass
class ModuleConfig:
    marker_inverts: dict[int, bool] = field(default_factory=dict)

    offset: OffsetConfig = field(default_factory=lambda: OffsetConfig())
    lo: LoConfig = field(default_factory=lambda: LoConfig())
    attenuation: AttConfig = field(default_factory=lambda: AttConfig())
    gain: GainConfig = field(default_factory=lambda: GainConfig())
    scope_acq: ScopeAcqConfig = field(default_factory=lambda: ScopeAcqConfig())


@dataclass
class QbloxConfig:
    slot_idx: int | None = None
    module: ModuleConfig = field(default_factory=lambda: ModuleConfig())
    sequencers: dict[int, SequencerConfig] = field(default_factory=dict)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from functools import partial
from inspect import signature
from typing import TYPE_CHECKING, Annotated, Generic, TypeVar

from pydantic import AfterValidator, ConfigDict, ImportString

from qat.instrument.base import InstrumentBuilder
from qat.utils.pydantic import NoExtraFieldsModel

if TYPE_CHECKING:
    from qat.core.pipelines.configurable import (
        ConfigurableCompilePipeline,
        ConfigurableExecutePipeline,
        ConfigurablePipeline,
    )
    from qat.core.pipelines.factory import PipelineFactory
    from qat.engines.native import NativeEngine
    from qat.model.hardware_model import PhysicalHardwareModel
    from qat.model.loaders.base import BaseModelLoader
    from qat.pipelines.pipeline import Pipeline
    from qat.pipelines.updateable import UpdateablePipeline
    from qat.purr.compiler.hardware_models import QuantumHardwareModel

from qat.core.config.validators import (
    is_backend,
    is_engine,
    is_frontend,
    is_hardwareloader,
    is_instrument,
    is_instrument_builder,
    is_middleend,
    is_passmanager_factory,
    is_pipeline_factory,
    is_pipeline_instance,
    is_runtime,
    is_target_data,
    is_updateable_pipeline,
)

ImportPipeline = Annotated[ImportString, AfterValidator(is_pipeline_instance)]
ImportPipelineFactory = Annotated[ImportString, AfterValidator(is_pipeline_factory)]
ImportUpdateablePipeline = Annotated[ImportString, AfterValidator(is_updateable_pipeline)]
ImportHardwareLoader = Annotated[ImportString, AfterValidator(is_hardwareloader)]

ImportFrontend = Annotated[ImportString, AfterValidator(is_frontend)]
ImportMiddleend = Annotated[ImportString, AfterValidator(is_middleend)]
ImportBackend = Annotated[ImportString, AfterValidator(is_backend)]
ImportEngine = Annotated[ImportString, AfterValidator(is_engine)]
ImportInstrument = Annotated[ImportString, AfterValidator(is_instrument)]
ImportInstrumentBuilder = Annotated[ImportString, AfterValidator(is_instrument_builder)]
ImportRuntime = Annotated[ImportString, AfterValidator(is_runtime)]
ImportTargetData = Annotated[ImportString, AfterValidator(is_target_data)]

ImportPassManagerFactory = Annotated[ImportString, AfterValidator(is_passmanager_factory)]


class PipelineInstanceDescription(NoExtraFieldsModel):
    name: str
    pipeline: ImportPipeline
    default: bool = False

    def construct(self) -> "Pipeline":
        """Returns the requested Pipeline instance"""
        return self.pipeline.copy_with_name(self.name)

    def is_subtype_of(self, ty: type) -> bool:
        """Matches the type of the pipeline against the given type."""
        return isinstance(self.pipeline, ty)


class HardwareLoaderDescription(NoExtraFieldsModel):
    name: str
    config: dict = {}
    type: ImportHardwareLoader

    def construct(self) -> "BaseModelLoader":
        """Returns the described Hardware Loader"""
        return self.type(**self.config)


class InstrumentBuilderDescription(NoExtraFieldsModel):
    name: str
    type: ImportInstrumentBuilder
    configs: list[dict] = []
    cinstr_type: ImportInstrument
    linstr_type: ImportInstrument

    def construct(self) -> "InstrumentBuilder":
        """Returns the described Instrument instance as an InstrumentBuilder."""
        return self.type(self.configs, self.cinstr_type, self.linstr_type)


class EngineDescription(NoExtraFieldsModel):
    name: str
    hardware_loader: str | None = None
    instrument_builder: InstrumentBuilderDescription | None = None
    config: dict = {}
    type: ImportEngine

    def construct(
        self,
        model: "None | QuantumHardwareModel | PhysicalHardwareModel" = None,
    ) -> "NativeEngine":
        """Returns the described Engine instance, injecting the model if provided."""
        kwargs = dict(self.config)
        if model is not None:
            kwargs["model"] = model

        instrument_builder = (
            self.instrument_builder.construct() if self.instrument_builder else None
        )
        instrument = instrument_builder.build() if instrument_builder else None
        if instrument is not None:
            kwargs["instrument"] = instrument
        return self.type(**kwargs)


T = TypeVar("T")


class ClassDescription(NoExtraFieldsModel, Generic[T]):
    type: T
    config: dict = {}

    def partial(self):
        """Returns a partially configured class"""
        try:
            return partial(self.type, **self.config)
        except TypeError as t:
            raise ValueError(f"Validation error {str(t)}")


ToPartialValidator = AfterValidator(lambda v: v.partial())

FrontendDescription = (
    ImportFrontend | Annotated[ClassDescription[ImportFrontend], ToPartialValidator]
)

MiddleendDescription = (
    ImportMiddleend | Annotated[ClassDescription[ImportMiddleend], ToPartialValidator]
)

BackendDescription = (
    ImportBackend | Annotated[ClassDescription[ImportBackend], ToPartialValidator]
)

RuntimeDescription = (
    ImportRuntime | Annotated[ClassDescription[ImportRuntime], ToPartialValidator]
)


TargetDataDescription = (
    ImportTargetData | Annotated[ClassDescription[ImportTargetData], ToPartialValidator]
)


PassManagerFactoryDescription = (
    ImportPassManagerFactory
    | Annotated[ClassDescription[ImportPassManagerFactory], ToPartialValidator]
)


class PipelineFactoryDescription(NoExtraFieldsModel):
    """A description pointing to a function that produces a pipeline, which is configured by
    a model, target data, an engine, and custom configuration."""

    name: str
    hardware_loader: str | None = None
    engine: str | None = None
    target_data: TargetDataDescription | None = None
    pipeline: ImportPipelineFactory
    config: dict = {}
    default: bool = False

    def construct(
        self, loader: "BaseModelLoader", engine: "NativeEngine | None" = None
    ) -> "PipelineFactory":
        """Constructs and returns a Pipeline from its description

        :param loader: The hardware model loader to fetch the hardware model.
        :param engine: The engine to use for the pipeline, optional.
        :return: The constructed pipeline
        """

        from qat.core.pipelines.factory import PipelineFactory

        target_data = self.target_data() if self.target_data is not None else None
        return PipelineFactory(
            config=self, loader=loader, target_data=target_data, engine=engine
        )

    def is_subtype_of(self, ty: type) -> bool:
        """Matches the type of the pipeline returned by the factory against the given
        type."""
        return issubclass(signature(self.pipeline).return_annotation, ty)


class UpdateablePipelineDescription(NoExtraFieldsModel):
    """A description pointing to an updateable pipeline, which can be configured with
    a custom hardware model (loader), target data, and an engine. It also always custom
    configuration (given by the concrete updateable pipeline class)."""

    name: str
    pipeline: ImportUpdateablePipeline
    hardware_loader: str | None = None
    engine: str | None = None
    target_data: TargetDataDescription | None = None
    config: dict = {}
    default: bool = False

    def construct(
        self, loader: "BaseModelLoader", engine: "NativeEngine | None" = None
    ) -> "UpdateablePipeline":
        """Constructs and returns a Pipeline from its description

        :param loader: The hardware model loader to fetch the hardware model.
        :param engine: The engine to use for the pipeline.
        :return: The updateable pipeline with a constructed pipeline instance.
        """

        config = self.config
        config["name"] = self.name

        target_data = self.target_data() if self.target_data is not None else None
        return self.pipeline(
            config=config, loader=loader, target_data=target_data, engine=engine
        )

    def is_subtype_of(self, ty: type) -> bool:
        """Matches the type of the pipeline returned by the factory against the given
        type."""
        return issubclass(signature(self.pipeline._build_pipeline).return_annotation, ty)


class CompilePipelineDescription(NoExtraFieldsModel):
    """A description of a compile pipeline, which is a pipeline that can be used to compile
    a model into an executable."""

    name: str
    hardware_loader: str | None = None
    frontend: FrontendDescription = "qat.frontend.DefaultFrontend"
    middleend: MiddleendDescription = "qat.middleend.DefaultMiddleend"
    backend: BackendDescription = "qat.backend.DefaultBackend"
    target_data: TargetDataDescription = "qat.model.target_data.DefaultTargetData"
    default: bool = False
    model_config = ConfigDict(validate_default=True)

    @staticmethod
    def is_subtype_of(ty: type) -> bool:
        """Matches the type of the pipeline returned by the factory against the given
        type."""
        from qat.pipelines.pipeline import CompilePipeline

        return issubclass(CompilePipeline, ty)

    def construct(
        self, loader: "BaseModelLoader", engine: "NativeEngine | None" = None
    ) -> "ConfigurableCompilePipeline":
        """Constructs and returns a CompilePipeline from its description

        :param loader: The hardware model loader to fetch the hardware model.
        :return: The constructed pipeline
        """

        from qat.core.pipelines.configurable import ConfigurableCompilePipeline

        target_data = self.target_data() if self.target_data is not None else None
        return ConfigurableCompilePipeline(
            config=self, loader=loader, target_data=target_data, engine=None
        )


class ExecutePipelineDescription(NoExtraFieldsModel):
    """A description of an executable pipeline, which is a pipeline that can be used to
    execute an executable."""

    name: str
    hardware_loader: str | None = None
    engine: str | None = None
    runtime: RuntimeDescription = "qat.runtime.DefaultRuntime"
    target_data: TargetDataDescription = "qat.model.target_data.DefaultTargetData"
    results_pipeline: PassManagerFactoryDescription = (
        "qat.runtime.results_pipeline.get_default_results_pipeline"
    )
    default: bool = False
    model_config = ConfigDict(validate_default=True)

    @staticmethod
    def is_subtype_of(ty: type) -> bool:
        """Matches the type of the pipeline returned by the factory against the given
        type."""
        from qat.pipelines.pipeline import ExecutePipeline

        return issubclass(ExecutePipeline, ty)

    def construct(
        self, loader: "BaseModelLoader", engine: "NativeEngine"
    ) -> "ConfigurableExecutePipeline":
        """Constructs and returns a Pipeline from its description

        :param loader: The hardware model loader to fetch the hardware model.
        :param engine: The engine to use for the pipeline.
        :return: The constructed pipeline
        """

        from qat.core.pipelines.configurable import ConfigurableExecutePipeline

        target_data = self.target_data() if self.target_data is not None else None
        return ConfigurableExecutePipeline(
            config=self, loader=loader, target_data=target_data, engine=engine
        )


class PipelineClassDescription(NoExtraFieldsModel):
    """Allows pipelines to be specified in a granular way and constructed as an updateable
    pipeline, allowing the hardware model to be refreshed."""

    name: str
    hardware_loader: str | None = None
    engine: str | None = None
    frontend: FrontendDescription = "qat.frontend.DefaultFrontend"
    middleend: MiddleendDescription = "qat.middleend.DefaultMiddleend"
    backend: BackendDescription = "qat.backend.DefaultBackend"
    runtime: RuntimeDescription = "qat.runtime.DefaultRuntime"
    target_data: TargetDataDescription = "qat.model.target_data.DefaultTargetData"
    results_pipeline: PassManagerFactoryDescription = (
        "qat.runtime.results_pipeline.get_default_results_pipeline"
    )
    default: bool = False
    model_config = ConfigDict(validate_default=True)

    def construct(
        self, loader: "BaseModelLoader", engine: "NativeEngine | None" = None
    ) -> "ConfigurablePipeline":
        """Constructs and returns a Pipeline from its description

        :param loader: The hardware model loader to fetch the hardware model.
        :param engine: The engine to use for the pipeline.
        :return: The pipeline as a :class:`ConfigurablePipeline` instance.
        """

        from qat.core.pipelines.configurable import ConfigurablePipeline
        from qat.engines import DefaultEngine

        if engine is None:
            engine = DefaultEngine()

        return ConfigurablePipeline(
            config=self, loader=loader, target_data=self.target_data(), engine=engine
        )

    @staticmethod
    def is_subtype_of(ty: type) -> bool:
        """Matches the type of the pipeline returned by the factory against the given
        type."""
        from qat.pipelines.pipeline import Pipeline

        return issubclass(Pipeline, ty)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from compiler_config.config import CompilerConfig

from qat.core.metrics_base import MetricsManager
from qat.core.pass_base import PassManager
from qat.core.result_base import ResultManager
from qat.purr.backends.live import LiveDeviceEngine
from qat.purr.backends.qiskit_simulator import QiskitBuilder
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.execution import InstructionExecutionEngine
from qat.purr.compiler.instructions import Repeat
from qat.runtime import BaseRuntime
from qat.runtime.connection import ConnectionMode


class LegacyRuntime(BaseRuntime):
    """A runtime that provides a wrapper around legacy engines for compatibility with the
    new stack."""

    def __init__(
        self,
        engine: InstructionExecutionEngine,
        results_pipeline: PassManager | None = None,
        connection_mode: ConnectionMode = ConnectionMode.DEFAULT,
    ):
        """
        :param engine: The execution engine for a target machine.
        :param results_pipeline: Optionally provided a pipeline for results processing. If
            not provided, a default pipeline is provided.
        :param connection_mode: Specifies how the connection is maintained.
        """
        if not isinstance(engine, InstructionExecutionEngine):
            raise ValueError(
                f"The engine provided has type {type(engine)}, which is not a "
                "`qat.purr.complier.execution.InstructionExecutionEngine`. Please use the "
                "`SimpleRuntime` for the refactored engines."
            )
        self.engine = engine
        self.connection_mode = connection_mode
        self.connect_engine(ConnectionMode.CONNECT_AT_BEGINNING)

        if not results_pipeline:
            results_pipeline = PassManager()
        self.results_pipeline = results_pipeline

    def execute(
        self,
        package: InstructionBuilder,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        **kwargs,
    ):
        """Fully execute QatIR against the hardware using a legacy execution engines.

        :param package: The program as an instruction builder.
        :param res_mgr: Optionally provide a results manager to save pass information.
        :param met_mgr: Optionally provide a metric manager to save pass information.
        :returns: Execution results.
        """

        if res_mgr is None:
            res_mgr = ResultManager()
        if met_mgr is None:
            met_mgr = MetricsManager()

        self.connect_engine(ConnectionMode.CONNECT_BEFORE_EXECUTE)
        acquisitions = self.engine.execute(package.instructions)
        self.disconnect_engine(ConnectionMode.DISCONNECT_AFTER_EXECUTE)

        package.shots = self._determine_shots(package, kwargs.get("compiler_config"))

        return self.results_pipeline.run(
            acquisitions, res_mgr, met_mgr, package=package, **kwargs
        )

    def _determine_shots(
        self, package: InstructionBuilder, compiler_config: CompilerConfig | None
    ) -> int:
        """Determine the number of shots to use for execution.

        :param package: The "package" that is being executed. For a
            :class:`LegacyRuntime` that is a Builder.
        :compiler_config: The config information for the task.
        """
        if hasattr(package, "shots") and package.shots is not None:
            return package.shots
        elif (
            isinstance(insts := package.instructions, QiskitBuilder)
            and insts.shot_count is not None
        ):
            return insts.shot_count
        elif (
            repeat_inst := next(filter(lambda x: isinstance(x, Repeat), insts), None)
        ) is not None and repeat_inst.repeat_count is not None:
            return repeat_inst.repeat_count
        elif compiler_config is not None and compiler_config.repeats is not None:
            return compiler_config.repeats
        else:
            return None

    def connect_engine(self, flag: ConnectionMode):
        """Connect the engine according to the connection mode."""
        if not isinstance(self.engine, LiveDeviceEngine):
            return

        if flag in self.connection_mode:
            self.engine.startup()

    def disconnect_engine(self, flag: ConnectionMode):
        """Disconnect the engine according to the connection mode."""
        if not isinstance(self.engine, LiveDeviceEngine):
            return

        if flag in self.connection_mode:
            self.engine.shutdown()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023-2025 Oxford Quantum Circuits Ltd
from dataclasses import dataclass
from typing import Dict

from qat.purr.compiler.devices import PulseShapeType
from qat.purr.compiler.instructions import Pulse


@dataclass
class WaveformDef:
    name: str
    type: str = "Unknown"
    description: str = "Empty"


class AbstractWaveform(Pulse):
    actual_waveforms: Dict[str, "AbstractWaveform"] = dict()
    waveform_definition: WaveformDef = None

    def __init_subclass__(cls: "AbstractWaveform"):
        if cls.waveform_definition is None:
            raise ValueError(
                f"Subclass of {AbstractWaveform.__name__} doesn't have a "
                "waveform_definition."
            )

        normalized_name = cls.waveform_definition.name.lower().replace(" ", "_")
        AbstractWaveform.actual_waveforms[normalized_name] = cls


class SquareWaveform(AbstractWaveform):
    waveform_definition: WaveformDef = WaveformDef(name="Square")

    def __init__(self, channel, width, amp, ignore_channel_scale):
        super().__init__(
            quantum_target=channel,
            shape=PulseShapeType.SQUARE,
            width=width,
            amp=amp,
            ignore_channel_scale=ignore_channel_scale,
        )


class GaussianWaveform(AbstractWaveform):
    waveform_definition: WaveformDef = WaveformDef(name="Gaussian")

    def __init__(self, channel, width, amp, ignore_channel_scale):
        super().__init__(
            quantum_target=channel,
            shape=PulseShapeType.GAUSSIAN,
            width=width,
            amp=amp,
            ignore_channel_scale=ignore_channel_scale,
        )


class SoftSquareWaveform(AbstractWaveform):
    waveform_definition: WaveformDef = WaveformDef(name="Soft square")

    def __init__(self, channel, width, amp, ignore_channel_scale):
        super().__init__(
            quantum_target=channel,
            shape=PulseShapeType.SOFT_SQUARE,
            width=width,
            amp=amp,
            ignore_channel_scale=ignore_channel_scale,
        )


class BlackmanWaveform(AbstractWaveform):
    waveform_definition: WaveformDef = WaveformDef(name="Blackman")

    def __init__(self, channel, width, amp, ignore_channel_scale):
        super().__init__(
            quantum_target=channel,
            shape=PulseShapeType.BLACKMAN,
            width=width,
            amp=amp,
            ignore_channel_scale=ignore_channel_scale,
        )


class SetupHoldWaveform(AbstractWaveform):
    waveform_definition: WaveformDef = WaveformDef(name="Setup hold")

    def __init__(self, channel, width, amp, ignore_channel_scale):
        super().__init__(
            quantum_target=channel,
            shape=PulseShapeType.SETUP_HOLD,
            width=width,
            amp=amp,
            ignore_channel_scale=ignore_channel_scale,
        )


class SofterSquareWaveform(AbstractWaveform):
    waveform_definition: WaveformDef = WaveformDef(name="Softer square")

    def __init__(self, channel, width, amp, ignore_channel_scale):
        super().__init__(
            quantum_target=channel,
            shape=PulseShapeType.SOFTER_SQUARE,
            width=width,
            amp=amp,
            ignore_channel_scale=ignore_channel_scale,
        )


class ExtraSoftSquareWaveform(AbstractWaveform):
    waveform_definition: WaveformDef = WaveformDef(name="Extra soft square")

    def __init__(self, channel, width, amp, ignore_channel_scale):
        super().__init__(
            quantum_target=channel,
            shape=PulseShapeType.EXTRA_SOFT_SQUARE,
            width=width,
            amp=amp,
            ignore_channel_scale=ignore_channel_scale,
        )


class SofterGaussianWaveform(AbstractWaveform):
    waveform_definition: WaveformDef = WaveformDef(name="Softer Gaussian")

    def __init__(self, channel, width, amp, ignore_channel_scale):
        super().__init__(
            quantum_target=channel,
            shape=PulseShapeType.SOFTER_GAUSSIAN,
            width=width,
            amp=amp,
            ignore_channel_scale=ignore_channel_scale,
        )


def get_waveform_type(wf_name: str):
    return AbstractWaveform.actual_waveforms.get(wf_name, None)


def build_waveform(wf_name: str, *args, **kwargs):
    wf = AbstractWaveform.actual_waveforms.get(wf_name, None)
    if wf is not None:
        return wf(*args, **kwargs)
    return None



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from dataclasses import dataclass
from enum import Enum
from typing import Any

import numpy as np


class Opcode(Enum):
    # Control
    NOP = "nop"
    STOP = "stop"

    # Jumps
    ADDRESS = ""
    JUMP = "jmp"
    JUMP_GREATER_EQUALS = "jge"
    JUMP_LESS_THAN = "jlt"
    LOOP = "loop"

    # Arithmetic
    MOVE = "move"
    ADD = "add"
    SUB = "sub"
    NOT = "not"
    AND = "and"
    OR = "or"
    XOR = "xor"

    # Parameter operation
    SET_MARKER = "set_mrk"
    SET_NCO_FREQUENCY = "set_freq"
    SET_NCO_PHASE = "set_ph"
    SET_NCO_PHASE_OFFSET = "set_ph_delta"
    RESET_PHASE = "reset_ph"
    SET_AWG_GAIN = "set_awg_gain"
    SET_AWG_OFFSET = "set_awg_offs"

    # Real-time pipeline
    SET_COND = "set_cond"
    UPDATE_PARAMETERS = "upd_param"
    PLAY = "play"
    ACQUIRE = "acquire"
    ACQUIRE_WEIGHED = "acquire_weighed"
    ACQUIRE_TTL = "acquire_ttl"
    SET_LATCH_EN = "set_latch_en"
    LATCH_RST = "latch_rst"
    WAIT = "wait"
    WAIT_SYNC = "wait_sync"
    WAIT_TRIGGER = "wait_trigger"


class Q1asmInstruction:
    def __init__(self, opcode: Opcode, *operands, comment: str = None):
        self.opcode: Opcode = opcode
        self.operands: tuple[Any, ...] = operands
        self.comment: str = comment

    def __str__(self):
        args = ",".join([str(op) for op in self.operands])

        if self.opcode == Opcode.ADDRESS:
            asm_str = f"{args}:"
        elif args:
            asm_str = f"{self.opcode.value} {args}"
        else:
            asm_str = f"{self.opcode.value}"

        if self.comment:
            asm_str += f" # {self.comment}"

        return asm_str

    def __repr__(self):
        return self.__str__()


@dataclass
class Sequence:
    waveforms: dict[str, dict[str, Any]]
    weights: dict[str, Any]
    acquisitions: dict[str, Any]
    program: str


class SequenceBuilder:
    """
    High-level builder apis for sequence codegen.
    This includes q1asm, waveforms, weights, and acquisitions.
    """

    def __init__(self):
        self.waveforms: dict[str, dict[str, Any]] = {}
        self.acquisitions: dict[str, Any] = {}
        self.weights: dict[str, Any] = {}
        self.q1asm_instructions: list[Q1asmInstruction] = []

    def build(self):
        return Sequence(
            waveforms=self.waveforms,
            acquisitions=self.acquisitions,
            weights=self.weights,
            program="\n".join([str(inst) for inst in self.q1asm_instructions]),
        )

    def lookup_waveform_by_data(self, data: np.ndarray):
        return next(
            (
                wf["index"]
                for wf in self.waveforms.values()
                if np.array(wf["data"]).shape == data.shape
                and np.all(np.array(wf["data"]) == data)
            ),
            None,
        )

    def add_waveform(self, name: str, index: int, data: list[float]):
        if name in self.waveforms:
            raise ValueError(f"A waveform named {name} already exists")
        self.waveforms[name] = {"index": index, "data": data}

    def add_acquisition(self, name: str, index: int, num_bins: int):
        if name in self.acquisitions:
            raise ValueError(f"An acquisition named {name} already exists")
        self.acquisitions[name] = {"index": index, "num_bins": num_bins}

    def add_weight(self, name: str, index: int, data: list[float]):
        if name in self.weights:
            raise ValueError(f"A weight named {name} already exists")
        self.weights[name] = {"index": index, "data": data}

    def _add_instruction(self, opcode: Opcode, *operands, comment: str = None):
        self.q1asm_instructions.append(Q1asmInstruction(opcode, *operands, comment=comment))
        return self

    # Control Instructions
    def nop(self, comment: str = None):
        return self._add_instruction(Opcode.NOP, comment=comment)

    def stop(self, comment: str = None):
        return self._add_instruction(Opcode.STOP, comment=comment)

    # Jump Instructions
    def label(self, label: int | str, comment: str = None):
        return self._add_instruction(Opcode.ADDRESS, label, comment=comment)

    def jmp(self, address: int | str, comment: str = None):
        if isinstance(address, str):
            address = f"@{address}"
        return self._add_instruction(Opcode.JUMP, address, comment=comment)

    def jge(self, a: str, b: int, address: int | str, comment: str = None):
        if isinstance(address, str):
            address = f"@{address}"
        return self._add_instruction(
            Opcode.JUMP_GREATER_EQUALS, a, b, address, comment=comment
        )

    def jlt(self, a: str, b: int, address: int | str, comment: str = None):
        if isinstance(address, str):
            address = f"@{address}"
        return self._add_instruction(Opcode.JUMP_LESS_THAN, a, b, address, comment=comment)

    def loop(self, a: str, address: int | str, comment: str = None):
        if isinstance(address, str):
            address = f"@{address}"
        return self._add_instruction(Opcode.LOOP, a, address, comment=comment)

    # Arithmetic instructions
    def move(self, src: int | str, dest: str, comment: str = None):
        return self._add_instruction(Opcode.MOVE, src, dest, comment=comment)

    def add(self, a: str, b: int | str, dest: str, comment: str = None):
        return self._add_instruction(Opcode.ADD, a, b, dest, comment=comment)

    def sub(self, a: str, b: int | str, dest: str, comment: str = None):
        return self._add_instruction(Opcode.SUB, a, b, dest, comment=comment)

    def logic_not(self, src: int | str, dest: str, comment: str = None):
        return self._add_instruction(Opcode.NOT, src, dest, comment=comment)

    def logic_and(self, a: str, b: int | str, dest: str, comment: str = None):
        return self._add_instruction(Opcode.AND, a, b, dest, comment=comment)

    def logic_or(self, a: str, b: int | str, dest: str, comment: str = None):
        return self._add_instruction(Opcode.OR, a, b, dest, comment=comment)

    def logic_xor(self, a: str, b: int | str, dest: str, comment: str = None):
        return self._add_instruction(Opcode.XOR, a, b, dest, comment=comment)

    # Parameter operation instructions
    def set_mrk(self, mask: int | str, comment: str = None):
        return self._add_instruction(Opcode.SET_MARKER, mask, comment=comment)

    def set_freq(self, frequency: int | str, comment: str = None):
        return self._add_instruction(Opcode.SET_NCO_FREQUENCY, frequency, comment=comment)

    def set_ph(self, phase: int | str, comment: str = None):
        return self._add_instruction(Opcode.SET_NCO_PHASE, phase, comment=comment)

    def set_ph_delta(self, phase_delta: int | str, comment: str = None):
        return self._add_instruction(
            Opcode.SET_NCO_PHASE_OFFSET, phase_delta, comment=comment
        )

    def reset_ph(self, comment: str = None):
        return self._add_instruction(Opcode.RESET_PHASE, comment=comment)

    def set_awg_gain(self, gain_i: int | str, gain_q: int | str, comment: str = None):
        return self._add_instruction(Opcode.SET_AWG_GAIN, gain_i, gain_q, comment=comment)

    def set_awg_offs(self, offset_i: int | str, offset_q: int | str, comment: str = None):
        return self._add_instruction(
            Opcode.SET_AWG_OFFSET, offset_i, offset_q, comment=comment
        )

    # Real-time instructions
    def set_cond(
        self,
        enable: int | str,
        mask: int | str,
        operator: int | str,
        else_duration: int,
        comment: str = None,
    ):
        return self._add_instruction(
            Opcode.SET_COND, enable, mask, operator, else_duration, comment=comment
        )

    def upd_param(self, duration: int, comment: str = None):
        return self._add_instruction(Opcode.UPDATE_PARAMETERS, duration, comment=comment)

    def play(
        self,
        wave_i: int | str,
        wave_q: int | str,
        duration: int,
        comment: str = None,
    ):
        return self._add_instruction(Opcode.PLAY, wave_i, wave_q, duration, comment=comment)

    def acquire(self, acq: int, bin: int | str, duration: int, comment: str = None):
        return self._add_instruction(Opcode.ACQUIRE, acq, bin, duration, comment=comment)

    def acquire_weighed(
        self,
        acq: int,
        bin: int | str,
        weight_i: int | str,
        weight_q: int | str,
        duration: int,
        comment: str = None,
    ):
        return self._add_instruction(
            Opcode.ACQUIRE_WEIGHED, acq, bin, weight_i, weight_q, duration, comment=comment
        )

    def acquire_ttl(
        self,
        acq: int,
        bin: int | str,
        enable: int,
        duration: int,
        comment: str = None,
    ):
        return self._add_instruction(
            Opcode.ACQUIRE_TTL, acq, bin, enable, duration, comment=comment
        )

    def set_latch_en(self, enable: int | str, duration: int, comment: str = None):
        return self._add_instruction(Opcode.SET_LATCH_EN, enable, duration, comment=comment)

    def latch_rst(self, duration: int | str, comment: str = None):
        return self._add_instruction(Opcode.LATCH_RST, duration, comment=comment)

    def wait(self, duration: int | str, comment: str = None):
        return self._add_instruction(Opcode.WAIT, duration, comment=comment)

    def wait_trigger(self, trigger: int | str, duration: int | str, comment: str = None):
        return self._add_instruction(
            Opcode.WAIT_TRIGGER, trigger, duration, comment=comment
        )

    def wait_sync(self, duration: int | str, comment: str = None):
        return self._add_instruction(Opcode.WAIT_SYNC, duration, comment=comment)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from compiler_config.config import CompilerConfig

from qat.core.metrics_base import MetricsManager
from qat.core.pass_base import PassManager
from qat.core.result_base import ResultManager
from qat.frontend.base import BaseFrontend
from qat.model.hardware_model import PhysicalHardwareModel as PydHardwareModel
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.qat import QATInput


class CustomFrontend(BaseFrontend):
    """
    Frontend that uses a custom pipeline to compile the input to an IR.
    While it is not equipped with a specific parser, it allows the user
    to specify custom compilation requirements via a pipeline.
    """

    # TODO: Allow the custom frontend to be equipped with an optional parser. Requires
    # the move over to pydantic hardware parsers with a shared API. (COMPILER-334)

    def __init__(
        self,
        model: None | QuantumHardwareModel | PydHardwareModel,
        pipeline: None | PassManager = None,
    ):
        """
        :param model: The hardware model that holds calibrated information on the qubits on the QPU,
                    defaults to None.
        :param pipeline: The custom pipeline, defaults to None.
        """
        super().__init__(model)
        self.pipeline = pipeline

    def check_and_return_source(self, src):
        """Custom frontends are too flexible to have any type checking, so just return the
        source file."""
        return src

    def emit(
        self,
        src: QATInput,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ):
        """
        Compiles an input :class:`QatInput` down to :class:`QatIR` with the custom
        pipeline and emits it.
        :param src: The high-level input.
        :param res_mgr: Collection of analysis results with caching and aggregation
                        capabilities, defaults to None.
        :param met_mgr: Stores useful intermediary metrics that are generated during
                        compilation, defaults to None.
        :param compiler_config: Compiler settings, defaults to None.
        :return: An intermediate representation as an :class:`InstructionBuilder` which
                holds a list of instructions to be executed on the QPU.
        """

        res_mgr = res_mgr or ResultManager()
        met_mgr = met_mgr or MetricsManager()
        compiler_config = compiler_config or CompilerConfig()

        ir = self.pipeline.run(
            src, res_mgr, met_mgr, compiler_config=compiler_config, **kwargs
        )
        return ir



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd


class ExecutionException(Exception):
    """Base class for exceptions for execution of programs in QAT."""

    ...



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import inspect
from typing import Callable

from qat.instrument.base import InstrumentBuilder, InstrumentConcept


def is_pipeline_instance(value: type):
    """A validator which raises when the input not a BasePipeline instance."""
    from qat.pipelines.pipeline import BasePipeline

    if not isinstance(value, BasePipeline):
        raise ValueError(f"{value} is not a valid Pipeline instance")
    return value


def is_frontend(value: type):
    """A validator which raises when the input not a Frontend."""
    from qat.frontend.base import BaseFrontend

    if not (inspect.isclass(value) and issubclass(value, BaseFrontend)):
        raise ValueError(f"{value} is not a valid Frontend")
    return value


def is_middleend(value: type):
    """A validator which raises when the input not a Middleend."""

    from qat.middleend.base import BaseMiddleend

    if not (inspect.isclass(value) and issubclass(value, BaseMiddleend)):
        raise ValueError(f"{value} is not a valid Middleend")
    return value


def is_backend(value: type):
    """A validator which raises when the input not a Backend."""

    from qat.backend.base import BaseBackend

    if not (inspect.isclass(value) and issubclass(value, BaseBackend)):
        raise ValueError(f"{value} is not a valid Backend")
    return value


def is_runtime(value: type):
    """A validator which raises when the input not a Runtime."""
    from qat.runtime import BaseRuntime

    if not (inspect.isclass(value) and issubclass(value, BaseRuntime)):
        raise ValueError(f"{value} is not a valid Runtime")
    return value


def is_instrument(value: type):
    """
    A validator which raises when the input not an Instrument.
    """

    if not (inspect.isclass(value) and issubclass(value, InstrumentConcept)):
        raise ValueError(f"{value} is not a valid Instrument builder")
    return value


def is_instrument_builder(value: type):
    """
    A validator which raises when the input not an Instrument builder.
    """

    if not (inspect.isclass(value) and issubclass(value, InstrumentBuilder)):
        raise ValueError(f"{value} is not a valid Instrument builder")
    return value


def is_engine(value: type):
    """A validator which raises when the input not an Engine.

    This permits NativeEngines and (legacy qat.purr) InstructionExecutionEngine.
    In future we intend to only accept NativeEngines (once qat.purr is deprecated)
    """
    from qat.engines import NativeEngine
    from qat.purr.compiler.execution import InstructionExecutionEngine

    if not (
        inspect.isclass(value)
        and issubclass(value, (NativeEngine, InstructionExecutionEngine))
    ):
        raise ValueError(
            f"{value} is not a valid Engine (either NativeEngine or legacy "
            "InstructionExecutionEngine)"
        )
    return value


def is_target_data(value: type):
    """A validator which raises when the input not a TargetData. Allows both TargetData
    classes and functions which return TargetData instances."""

    from qat.model.target_data import AbstractTargetData

    if inspect.isclass(value) and issubclass(value, AbstractTargetData):
        return value

    if callable(value):
        return_annotation = inspect.signature(value).return_annotation
        if isinstance(return_annotation, type) and issubclass(
            return_annotation, AbstractTargetData
        ):
            return value
    raise ValueError(f"{value} is not a valid TargetData")


def is_passmanager(value: type):
    """A validator which raises when the input not an PassManager."""
    from qat.core.pass_base import PassManager

    if not (inspect.isclass(value) and issubclass(value, PassManager)):
        raise ValueError(f"{value} is not a valid PassManager")
    return value


def is_passmanager_factory(value: type):
    """A validator which raises when the input is not function that returns a
    PassManager."""
    from qat.core.pass_base import PassManager

    check_callable(value, PassManager)
    return value


def is_hardwareloader(value: type):
    """A validator which raises when the input not a model loader."""
    from qat.model.loaders.base import BaseModelLoader

    if not (inspect.isclass(value) and issubclass(value, BaseModelLoader)):
        raise ValueError(f"{value} is not a valid Hardware Model Loader")
    return value


def is_pipeline_factory(value: type):
    """A validator which raises when the input not function which returns a Pipeline."""
    from qat.pipelines.base import BasePipeline

    check_callable(value, BasePipeline)
    return value


def is_updateable_pipeline(value: type):
    """A validator which raises when the input not an UpdateablePipeline instance."""
    from qat.pipelines.updateable import UpdateablePipeline

    if not (inspect.isclass(value) and issubclass(value, UpdateablePipeline)):
        raise ValueError(f"{value} is not a valid UpdateablePipeline.")
    return value


def requires_model(value: type) -> bool:
    """Determines whether a class or function expects a model

    In future we will also inspect the type is a hardware model but currently
    we are not consistent enough with type hinting for this.
    """
    return "model" in inspect.signature(value).parameters


def requires_target_data(value: type) -> bool:
    """Determines whether a class or function expects target data.

    In future we will also inspect the type is a target data but currently we are not
    consistent enough with type hinting for this.
    """
    return "target_data" in inspect.signature(value).parameters


def check_callable(value: Callable, return_type: type):
    """Raises if the value is not a callable, and the return type is not the expected
    type."""

    if not callable(value):
        raise ValueError(f"{value} is not callable")

    return_annotation = inspect.signature(value).return_annotation
    if not isinstance(return_annotation, type):
        raise ValueError(f"{value} does not have a valid return annotation")

    if not issubclass(return_annotation, return_type):
        raise ValueError(f"{value} does not return a {return_type.__name__}")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from collections import defaultdict
from numbers import Number

import numpy as np
from compiler_config.config import CompilerConfig, QuantumResultsFormat, ResultsFormatting

from qat.core.config.configure import get_config
from qat.core.pass_base import ValidationPass
from qat.core.result_base import ResultManager
from qat.middleend.passes.purr.analysis import ActiveChannelResults
from qat.model.target_data import TargetData
from qat.purr.backends.live import LiveHardwareModel
from qat.purr.backends.qiskit_simulator import QiskitBuilder
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PulseChannel, Qubit
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import (
    Acquire,
    AcquireMode,
    CustomPulse,
    FrequencyShift,
    PostProcessing,
    ProcessAxis,
    Pulse,
    PulseShapeType,
    Repeat,
    Return,
    Sweep,
    Synchronize,
    Variable,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class PhysicalChannelAmplitudeValidation(ValidationPass):
    """
    Validates that the total amplitude of pulses on a physical channel does not exceed
    allowed levels.

    Data is stored in a nested dict structure

    .. code-block:: python

        { 'CH1': {
                      CH1.Q0.drive: 0.2+0.5j,
                      CH1.Q0.Q1.cross_resonsance: 0.4+0.0j,
                  },
          'CH2': { CH2.R0.measure: 0.9-0.6j },
        }

    This pass requires all :class:`Pulse` s to have been lowered to either :class:`PulseShapeType.SQUARE`
    or a :class:`CustomPulse` with the evaluated pulse `samples`. Any non lowered pulses with
    other types may cause the pipeline to fail.
    """

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """
        :param ir: The intermediate representation (IR) :class:`InstructionBuilder`.
        :raises ValueError: for :class:`Pulse` of non :class:`PulseShapeType.SQUARE` type.
        """
        # TODO: Evaluate if this would be better if operating on `PartitionedIR` class
        # instead of `InstructionBuilder`

        phys_chs_amps: dict[str, dict[str, complex]] = {}
        for pulse_channel in ir.model.pulse_channels.values():
            pulse_ch_amps = phys_chs_amps.setdefault(pulse_channel.physical_channel_id, {})
            pulse_ch_amps[pulse_channel] = 0.0

        for instruction in ir.instructions:
            if isinstance(instruction, Pulse | CustomPulse):
                pulse_ch_amps = phys_chs_amps[
                    phys_chan_id := instruction.channel.physical_channel_id
                ]
                pulse_chan = instruction.channel
                if isinstance(instruction, CustomPulse):
                    pulse_shape = np.array(instruction.samples)

                    # TODO: This may be part of the lowering or needed here.
                    # Needs to be determined and potentially adjusted as part of:
                    # COMPILER-???
                    # if not instruction.ignore_channel_scale:
                    #    pulse_shape *= instruction.channel.scale

                    pulse_ch_amps[pulse_chan] = complex(
                        max(np.abs(pulse_shape.real)) + 1j * max(np.abs(pulse_shape.imag))
                    )
                else:
                    if not instruction.shape == PulseShapeType.SQUARE:
                        raise ValueError(
                            f"Can not validate amplitude of un-lowered {instruction.shape} Pulse."
                        )
                    pulse_amp = complex(
                        instruction.scale_factor
                        * instruction.amp
                        * np.exp(1.0j * instruction.phase)
                    )
                    if not instruction.ignore_channel_scale:
                        pulse_amp *= instruction.channel.scale

                    pulse_ch_amps[pulse_chan] = pulse_amp

                total_amp = sum(pulse_ch_amps.values())

                if np.abs(np.real(total_amp)) > 1 or np.abs(np.imag(total_amp)) > 1:
                    raise ValueError(
                        f"Overflow error detect on {phys_chan_id}!"
                        " The sum of logical channels onto a physical channel is too high!"
                    )
            elif isinstance(instruction, Synchronize):
                handled_phys_chan_ids = []
                for target in instruction.quantum_targets:
                    phys_chan_id = target.physical_channel_id
                    if phys_chan_id in handled_phys_chan_ids:
                        continue

                    tmp = phys_chs_amps[phys_chan_id]
                    if all([chan in instruction.quantum_targets for chan in tmp.keys()]):
                        phys_chs_amps[phys_chan_id] = dict.fromkeys(tmp.keys(), 0.0)
                    handled_phys_chan_ids.append(phys_chan_id)

        return ir


class QiskitResultsFormatValidation(ValidationPass):
    """Validates the results format contains `BinaryCount`, and throws a warning if not."""

    def run(self, ir: QiskitBuilder, *args, compiler_config: CompilerConfig, **kwargs):
        """
        :param ir: The Qiskit instruction builder.
        :param compiler_config: The compiler config contains the results format.
        :return: The instruction builder, unaltered.
        """
        results_format = compiler_config.results_format
        format_flags = (
            results_format.transforms
            if isinstance(results_format, QuantumResultsFormat)
            else results_format
        )
        if format_flags is None or ResultsFormatting.BinaryCount not in format_flags:
            log.warning(
                "The results formatting `BinaryCount` was not found in the formatting "
                "flags. Please note that the Qiskit runtime only currently supports "
                "results returned as a binary count."
            )
        return ir


class InstructionValidation(ValidationPass):
    """Validates instructions against the hardware.

    Extracted from :mod:`qat.purr.compiler.execution.QuantumExecutionEngine.validate`.
    """

    def __init__(
        self,
        target_data: TargetData,
        pulse_duration_limits: bool | None = None,
        *args,
        **kwargs,
    ):
        """
        :param target_data: Target-related information.
        :param pulse_duration_limits: Whether to check the pulse duration limits.
            If None, uses the default from the QatConfig.
        :param instruction_memory_size: The maximum number of instructions that can
            be run in a single shot. If None, uses the default from QatConfig.
        """
        self.instruction_memory_size = target_data.instruction_memory_size
        self.pulse_duration_max = max(
            target_data.QUBIT_DATA.pulse_duration_max,
            target_data.RESONATOR_DATA.pulse_duration_max,
        )
        self.pulse_duration_limits = (
            pulse_duration_limits
            if pulse_duration_limits is not None
            else get_config().INSTRUCTION_VALIDATION.PULSE_DURATION_LIMITS
        )

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        instruction_length = len(ir.instructions)
        if instruction_length > self.instruction_memory_size:
            raise ValueError(
                f"Program with {instruction_length} instructions too large to be run in a single block on current hardware."
            )

        for inst in ir.instructions:
            if isinstance(inst, Acquire) and not inst.channel.acquire_allowed:
                raise ValueError(
                    f"Cannot perform an acquire on the physical channel with id "
                    f"{inst.channel.physical_channel}"
                )
            if isinstance(inst, (Pulse, CustomPulse)):
                duration = inst.duration
                if isinstance(duration, Number) and duration > self.pulse_duration_max:
                    if self.pulse_duration_limits:
                        # Do not throw error if we specifically disabled the limit checks.
                        # TODO: Add a lower bound for the pulse duration limits as well in a later PR,
                        # which is specific to each hardware model and can be stored as a member variables there.
                        raise ValueError(
                            f"Max Waveform width is {self.pulse_duration_max} s "
                            f"given: {inst.duration} s"
                        )
                elif isinstance(duration, Variable):
                    values = next(
                        iter(
                            [
                                sw.variables[duration.name]
                                for sw in ir.instructions
                                if isinstance(sw, Sweep)
                                and duration.name in sw.variables.keys()
                            ]
                        )
                    )
                    if np.max(values) > self.pulse_duration_max:
                        if self.pulse_duration_limits:
                            raise ValueError(
                                f"Max Waveform width is {self.pulse_duration_max} s "
                                f"given: {values} s"
                            )
        return ir


class ReadoutValidation(ValidationPass):
    """Validates that there are no mid-circuit measurements, and that the post-processing
    instructions do not have an invalid sequence.

    Extracted from :meth:`qat.purr.backends.live.LiveDeviceEngine.validate`.
    """

    # TODO: break this down into smaller passes!

    def __init__(
        self,
        hardware: QuantumHardwareModel,
        no_mid_circuit_measurement: bool | None = None,
        *args,
        **kwargs,
    ):
        """
        :param hardware: The hardware model is needed to check for mid-circuit measurments.
        :param no_mid_circuit_measurement: Whether mid-circuit measurements are allowed.
            If None, uses the default from the QatConfig.
        """
        self.hardware = hardware
        self.no_mid_circuit_measurement = (
            no_mid_circuit_measurement
            if no_mid_circuit_measurement is not None
            else get_config().INSTRUCTION_VALIDATION.NO_MID_CIRCUIT_MEASUREMENT
        )

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        # TODO -COMPILER-851 - Specify ctrl HW features in TargetData
        enable_hw_averaging = kwargs.get("enable_hw_averaging", False)

        model = self.hardware

        if not isinstance(model, LiveHardwareModel):
            return ir

        consumed_qubits: list[str] = []
        chanbits_map = {}
        for inst in ir.instructions:
            if isinstance(inst, PostProcessing):
                if (
                    inst.acquire.mode == AcquireMode.SCOPE
                    and ProcessAxis.SEQUENCE in inst.axes
                ):
                    if not enable_hw_averaging:
                        raise ValueError(
                            "Invalid post-processing! Post-processing over SEQUENCE is "
                            "not possible after the result is returned from hardware "
                            "in SCOPE mode!"
                        )
                elif (
                    inst.acquire.mode == AcquireMode.INTEGRATOR
                    and ProcessAxis.TIME in inst.axes
                ):
                    raise ValueError(
                        "Invalid post-processing! Post-processing over TIME is not "
                        "possible after the result is returned from hardware in "
                        "INTEGRATOR mode!"
                    )
                elif inst.acquire.mode == AcquireMode.RAW:
                    raise ValueError(
                        "Invalid acquire mode! The live hardware doesn't support "
                        "RAW acquire mode!"
                    )

            # Check if we've got a measure in the middle of the circuit somewhere.
            elif self.no_mid_circuit_measurement:
                if isinstance(inst, Acquire):
                    for qbit in model.qubits:
                        if qbit.get_acquire_channel() == inst.channel:
                            consumed_qubits.append(qbit)
                elif isinstance(inst, Pulse):
                    # Find target qubit from instruction and check whether it's been
                    # measured already.
                    acquired_qubits = [
                        (
                            (
                                chanbits_map[chanbit]
                                if chanbit in chanbits_map
                                else chanbits_map.setdefault(
                                    chanbit,
                                    model._resolve_qb_pulse_channel(chanbit)[0],
                                )
                            )
                            in consumed_qubits
                        )
                        for chanbit in inst.quantum_targets
                        if isinstance(chanbit, (Qubit, PulseChannel))
                    ]

                    if any(acquired_qubits):
                        raise ValueError(
                            "Mid-circuit measurements currently unable to be used."
                        )
        return ir


class HardwareConfigValidity(ValidationPass):
    """Validates the :class:`CompilerConfig` against the hardware model."""

    def __init__(self, hardware_model: QuantumHardwareModel):
        """Instantiate the pass with a hardware model.

        :param hardware_model: The hardware model.
        """
        self.hardware_model = hardware_model

    def run(
        self,
        ir: InstructionBuilder,
        *args,
        compiler_config: CompilerConfig,
        **kwargs,
    ):
        """
        :param compiler_config: The config containing compilation settings provided as a
            keyword argument.
        """
        compiler_config.validate(self.hardware_model)
        return ir


class FrequencySetupValidation(ValidationPass):
    """Validates the baseband frequencies and intermediate frequencies of pulse channels
    against the target data.
    """

    def __init__(self, model: QuantumHardwareModel, target_data: TargetData):
        """Instantiate the pass with a hardware model.

        :param model: The hardware model.
        :param target_data: Target-related information.
        """

        # Extract limits from the target data.
        self.qubit_lo_freq_limits = (
            target_data.QUBIT_DATA.pulse_channel_lo_freq_min,
            target_data.QUBIT_DATA.pulse_channel_lo_freq_max,
        )
        self.resonator_lo_freq_limits = (
            target_data.RESONATOR_DATA.pulse_channel_lo_freq_min,
            target_data.RESONATOR_DATA.pulse_channel_lo_freq_max,
        )
        self.qubit_if_freq_limits = (
            target_data.QUBIT_DATA.pulse_channel_if_freq_min,
            target_data.QUBIT_DATA.pulse_channel_if_freq_max,
        )
        self.resonator_if_freq_limits = (
            target_data.RESONATOR_DATA.pulse_channel_if_freq_min,
            target_data.RESONATOR_DATA.pulse_channel_if_freq_max,
        )

        # Calculate mappings for physical and pulse channels for quick lookup
        self._baseband_frequencies = self._create_baseband_frequency_map(model)
        self._is_resonator = self._create_resonator_map(model)
        self._pulse_channel_ifs = self._create_pulse_channel_if_map(model)
        self._pulse_to_physical_channel = (
            self._create_pulse_channel_to_physical_channel_map(model)
        )

        # Do validation prior to runtime
        self._baseband_valid = self._validate_baseband_frequencies(
            self._baseband_frequencies,
            self._is_resonator,
            self.qubit_lo_freq_limits,
            self.resonator_lo_freq_limits,
        )

        self._pulse_channels_valid = self._validate_pulse_channel_ifs(
            self._pulse_channel_ifs,
            self._is_resonator,
            self._pulse_to_physical_channel,
            self.qubit_if_freq_limits,
            self.resonator_if_freq_limits,
        )

    @staticmethod
    def _create_baseband_frequency_map(model: QuantumHardwareModel) -> dict[str, float]:
        """Creates a map of baseband frequencies for each physical channel in the model."""
        return {
            device.physical_channel.id: device.physical_channel.baseband_frequency
            for device in model.quantum_devices.values()
        }

    @staticmethod
    def _create_resonator_map(model: QuantumHardwareModel) -> dict[str, bool]:
        """Creates a map to lookup if physical channels belong to resonators.."""
        return {
            device.physical_channel.id: not isinstance(device, Qubit)
            for device in model.quantum_devices.values()
        }

    @staticmethod
    def _create_pulse_channel_if_map(model: QuantumHardwareModel) -> dict[str, float]:
        """Creates a map of pulse channel intermediate frequencies for each pulse channel
        in the model."""
        return {
            pulse_channel.partial_id(): (
                pulse_channel.frequency - pulse_channel.physical_channel.baseband_frequency
            )
            for pulse_channel in model.pulse_channels.values()
        }

    @staticmethod
    def _create_pulse_channel_to_physical_channel_map(
        model: QuantumHardwareModel,
    ) -> dict[str, str]:
        """Creates a map of pulse channels to their physical channels."""
        return {
            pulse_channel.partial_id(): pulse_channel.physical_channel.id
            for pulse_channel in model.pulse_channels.values()
        }

    @staticmethod
    def _validate_baseband_frequencies(
        frequencies: dict[str, float],
        is_resonator: dict[str, bool],
        qubit_lo_freq_limits: tuple[float, float],
        resonator_lo_freq_limits: tuple[float, float],
    ) -> dict[str, bool]:
        """Validates that the frequencies in the dictionary are within the specified range."""

        baseband_validation = {}
        for key, frequency in frequencies.items():
            lower, upper = (
                qubit_lo_freq_limits if not is_resonator[key] else resonator_lo_freq_limits
            )
            baseband_validation[key] = lower <= frequency <= upper
        return baseband_validation

    @staticmethod
    def _validate_pulse_channel_ifs(
        pulse_channel_ifs: dict[str, float],
        is_resonator: dict[str, bool],
        pulse_channel_to_physical_channel_map: dict[str, str],
        qubit_if_freq_limits: tuple[float, float],
        resonator_if_freq_limits: tuple[float, float],
    ) -> dict[str, bool]:
        """Validates that the pulse channel frequencies are within the specified range."""

        pulse_channel_validation = {}
        for pulse_channel, frequency in pulse_channel_ifs.items():
            lower, upper = (
                qubit_if_freq_limits
                if not is_resonator[pulse_channel_to_physical_channel_map[pulse_channel]]
                else resonator_if_freq_limits
            )
            pulse_channel_validation[pulse_channel] = bool(
                lower <= np.abs(frequency) <= upper
            )
        return pulse_channel_validation

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        """

        active_channel_res = res_mgr.lookup_by_type(ActiveChannelResults)
        pulse_channels = active_channel_res.targets
        physical_channels = set(
            [pulse_channel.physical_channel for pulse_channel in pulse_channels]
        )

        violations = []
        for physical_channel in physical_channels:
            if not self._baseband_valid[physical_channel.id]:
                freq_range = (
                    self.qubit_lo_freq_limits
                    if not self._is_resonator[physical_channel.id]
                    else self.resonator_lo_freq_limits
                )
                violations.append(
                    f"Physical channel {physical_channel.full_id()} with baseband frequency "
                    f"{self._baseband_frequencies[physical_channel.id]} is out of the valid "
                    f"range {freq_range}."
                )

        for pulse_channel in pulse_channels:
            if pulse_channel.partial_id() in self._pulse_channels_valid:
                valid = self._pulse_channels_valid[pulse_channel.partial_id()]
                if_freq = self._pulse_channel_ifs[pulse_channel.partial_id()]
            else:
                # We have to handle custom pulse channels manually
                if_freq = (
                    pulse_channel.frequency
                    - pulse_channel.physical_channel.baseband_frequency
                )
                if self._is_resonator[pulse_channel.physical_channel.id]:
                    lower, upper = self.resonator_if_freq_limits
                else:
                    lower, upper = self.qubit_if_freq_limits
                valid = bool(lower <= np.abs(if_freq) <= upper)

            if not valid:
                freq_range = (
                    self.qubit_if_freq_limits
                    if not self._is_resonator[pulse_channel.physical_channel.id]
                    else self.resonator_if_freq_limits
                )
                violations.append(
                    f"Pulse channel {pulse_channel.full_id()} with IF {if_freq} is out of "
                    f"the valid range {freq_range}."
                )

        if len(violations) > 0:
            raise ValueError(
                "Frequency validation of the hardware model against the target data failed "
                "with the following violations: \n" + "\n".join(violations)
            )
        return ir


class DynamicFrequencyValidation(ValidationPass):
    """Validates the setting or shifting frequencies does not move the intermediate
    frequency of a pulse channel outside the allowed limits."""

    def __init__(self, model: QuantumHardwareModel, target_data: TargetData):
        """Instantiate the pass with a hardware model.

        :param model: The hardware model.
        :param target_data: Target-related information.
        """
        self.model = model
        self._is_resonator = self._create_resonator_map(model)

        # Extract limits from the target data.
        self.qubit_if_freq_limits = (
            target_data.QUBIT_DATA.pulse_channel_if_freq_min,
            target_data.QUBIT_DATA.pulse_channel_if_freq_max,
        )
        self.resonator_if_freq_limits = (
            target_data.RESONATOR_DATA.pulse_channel_if_freq_min,
            target_data.RESONATOR_DATA.pulse_channel_if_freq_max,
        )

    @staticmethod
    def _create_resonator_map(model: QuantumHardwareModel) -> dict[str, bool]:
        """Creates a map to lookup if physical channels belong to resonators.."""
        return {
            device.physical_channel.id: not isinstance(device, Qubit)
            for device in model.quantum_devices.values()
        }

    def _validate_frequency_shifts(self, pulse_channel: PulseChannel, ifs: list[float]):
        """Validates that the frequency shifts do not exceed the allowed limits."""

        if_limits = (
            self.qubit_if_freq_limits
            if not self._is_resonator[pulse_channel.physical_channel.id]
            else self.resonator_if_freq_limits
        )

        return [
            if_value
            for if_value in ifs
            if not (if_limits[0] <= np.abs(if_value) <= if_limits[1])
        ]

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        ifs = defaultdict(list)
        for instruction in ir.instructions:
            if isinstance(instruction, FrequencyShift):
                shifts = ifs[instruction.channel]
                freq = (
                    shifts[-1]
                    if len(shifts) > 0
                    else instruction.channel.frequency
                    - instruction.channel.physical_channel.baseband_frequency
                )
                shifts.append(freq + instruction.frequency)
            # TODO: Add support for frequency set (COMPILER-644)

        violations = []
        for pulse_channel, if_values in ifs.items():
            if if_violations := self._validate_frequency_shifts(pulse_channel, if_values):
                violations.append(
                    f"The IF of pulse channel {pulse_channel.full_id()} is frequency "
                    f"shifted to values {if_violations} that exceed the allowed limits."
                )

        if len(violations) > 0:
            raise ValueError(
                "Dynamic frequency validation failed with the following violations:\n"
                + "\n".join(violations)
            )
        return ir


class FixedIntermediateFrequencyValidation(ValidationPass):
    """Checks that no frequency shifts or sets are applied to pulse channels that have
    a fixed intermediate frequency, or share a physical channel with a pulse channel."""

    def __init__(self, model: QuantumHardwareModel):
        """Instantiate the pass with a hardware model.

        :param model: The hardware model.
        """
        self.model = model
        self._fixed_ifs = self._create_fixed_if_map(model)

    @staticmethod
    def _create_fixed_if_map(model: QuantumHardwareModel) -> dict[str, bool]:
        """Creates a map of fixed intermediate frequencies for each pulse channel in the model."""
        fixed_ifs = {}
        for pulse_channel in model.pulse_channels.values():
            fixed_ifs[pulse_channel.physical_channel.id] = (
                pulse_channel.fixed_if
                or fixed_ifs.get(pulse_channel.physical_channel.id, False)
            )

        return fixed_ifs

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        violations = set()
        for instruction in ir.instructions:
            if isinstance(instruction, FrequencyShift):
                pulse_channel = instruction.channel
                if self._fixed_ifs[pulse_channel.physical_channel.id]:
                    violations.add(
                        f"Pulse channel {pulse_channel.full_id()} has a fixed IF and "
                        "cannot be frequency shifted."
                    )

        if len(violations) > 0:
            raise ValueError(
                "Fixed intermediate frequency validation failed with the following "
                "violations:\n".join(list(violations))
            )
        return ir


class ReturnSanitisationValidation(ValidationPass):
    """Validates that the IR has a :class:`Return` instruction."""

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        returns = [inst for inst in ir.instructions if isinstance(inst, Return)]

        if not returns:
            raise ValueError("Could not find any return instructions")
        elif len(returns) > 1:
            raise ValueError("Found multiple return instructions")
        return ir


class RepeatSanitisationValidation(ValidationPass):
    """Checks if the builder has a :class:`Repeat` instruction and warns if none exists."""

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        repeats = [inst for inst in ir.instructions if isinstance(inst, Repeat)]
        if not repeats:
            log.warning("Could not find any repeat instructions.")
        return ir



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from abc import ABC, abstractmethod

from qat.backend.graph import BasicBlock, ControlFlowGraph


class DfsTraversal(ABC):
    """
    Base Depth-first Search algorithm on the control flow graph
    """

    def __init__(self):
        self._entered: list[BasicBlock] = []

    def clear(self):
        self._entered.clear()

    def run(self, graph: ControlFlowGraph):
        self.clear()
        self._visit(graph.entry, graph)

    def _visit(self, node: BasicBlock, graph: ControlFlowGraph):
        self.enter(node)
        self._entered.append(node)
        for neighbour in graph.out_nbrs(node):
            if neighbour not in self._entered:
                self._visit(neighbour, graph)
        self.exit(node)

    @abstractmethod
    def enter(self, node: BasicBlock):
        pass

    @abstractmethod
    def exit(self, node: BasicBlock):
        pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from functools import singledispatchmethod

import numpy as np

from qat.ir.gates.base import GateBase
from qat.ir.gates.gates_1q import (
    Gate1Q,
    Hadamard,
    Id,
    Phase,
    Rx,
    Ry,
    Rz,
    S,
    Sdg,
    T,
    Tdg,
    U,
    X,
    Y,
    Z,
)
from qat.ir.gates.gates_2q import CNOT, ECR, SWAP, CPhase, CRx, CRy, CRz
from qat.ir.gates.native import NativeGate, X_pi_2, Z_phase, ZX_pi_4
from qat.middleend.decompositions.base import DecompositionBase


class GateDecompositionBase(DecompositionBase):
    """Base object for implementing decompositions of gates with :class:`NativeGate` for
    end nodes.

    This handles decompositions of gates to gates, and should not be used to lower gates to
    pulse instructions. While it is recommended to use :class:`DefaultGateDecompositons`,
    custom gate decompositions strategies can be implemented via this class. This could be
    used, for example, if targeting machines that have a different native gate set, or
    trying out approximate decompositions that have a cheaper implementation cost.
    """

    end_nodes = (NativeGate,)


class DefaultGateDecompositions(GateDecompositionBase):
    """Implements the standard gate decompositions used in OQC hardware.

    Provides decomposition rules for the native gate set {:class:`Z_phase`,
    :class:`X_pi_2`, :class:`ZX_pi_4`}.
    """

    end_nodes = (Z_phase, X_pi_2, ZX_pi_4)

    @singledispatchmethod
    def decompose_op(self, gate: GateBase):
        """Implements the definition of a decomposition of a gate.

        The definition does not have to be in terms of native gates, but decompositions
        must form a DAG. For example,

        .. code-block:: python

            CNOT -> {ECR, X, Z} -> {ZX_pi_4, Z_phase, X_pi_2}
            U -> {x_pi_2, Z_phase}
        """
        raise NotImplementedError(
            f"Decomposition for gate {gate.__class__.__name__} not implemented."
        )

    @decompose_op.register(Gate1Q)
    def _(self, gate: Gate1Q):
        """Decomposes a generic one qubit gate.

        A fallback method for when a decomposition is not implemented. Converts the gate to
        a :class:`U` gate, and uses that decomposition in its place.

        .. warning::
            Using this is likely to be less performant than a specific implementation. It is
            recommended to register a new decomposition for each new gate that is added.
        """

        new_gate = U.from_matrix(gate.qubit, gate.matrix)
        return self.decompose(new_gate)

    @decompose_op.register(U)
    def _(self, gate: U):
        """Decomposes the U gate into a product of X_pi_2 gates and Rz gates.

        Does not deal with special cases that can reduce the number of :class:`X_pi_2`
        gates: this is dealt with my optimisation passes.
        """

        return [
            Z_phase(qubit=gate.qubit, theta=gate.lambd + np.pi),
            X_pi_2(qubit=gate.qubit),
            Z_phase(qubit=gate.qubit, theta=np.pi - gate.theta),
            X_pi_2(qubit=gate.qubit),
            Z_phase(qubit=gate.qubit, theta=gate.phi),
        ]

    @decompose_op.register(X)
    def _(self, gate: X):
        """Decomposes the Pauli-X gate."""
        return [Rx(qubit=gate.qubit, theta=np.pi)]

    @decompose_op.register(Y)
    def _(self, gate: Y):
        """Decomposes the Pauli-Y gate."""
        return [Ry(qubit=gate.qubit, theta=np.pi)]

    @decompose_op.register(Z)
    def _(self, gate: Z):
        """Decomposes the Pauli-Z gate."""
        return [Rz(qubit=gate.qubit, theta=np.pi)]

    @decompose_op.register(Rx)
    def _(self, gate: Rx):
        """Decomposes the X-rotation gate.

        Does not deal with special cases that can reduce the number of :class:`X_pi_2`
        gates: this is dealt with my optimisation passes.
        """
        return [
            Z_phase(qubit=gate.qubit, theta=-np.pi / 2),
            X_pi_2(qubit=gate.qubit),
            Z_phase(qubit=gate.qubit, theta=np.pi - gate.theta),
            X_pi_2(qubit=gate.qubit),
            Z_phase(qubit=gate.qubit, theta=-np.pi / 2),
        ]

    @decompose_op.register(Ry)
    def _(self, gate: Ry):
        """Decomposes the Y-rotation gate.

        Does not deal with special cases that can reduce the number of :class:`X_pi_2`
        gates: this is dealt with my optimisation passes.
        """
        return [
            Z_phase(qubit=gate.qubit, theta=np.pi),
            X_pi_2(qubit=gate.qubit),
            Z_phase(qubit=gate.qubit, theta=np.pi - gate.theta),
            X_pi_2(qubit=gate.qubit),
        ]

    @decompose_op.register(Rz)
    def _(self, gate: Rz):
        """Decomposes the Z-rotation gate."""
        return [Z_phase(qubit=gate.qubit, theta=gate.theta)]

    @decompose_op.register(Phase)
    def _(self, gate: Phase):
        """Decomposes the Z-rotation gate."""
        return [Z_phase(qubit=gate.qubit, theta=gate.theta)]

    @decompose_op.register(Id)
    def _(self, _: Id):
        """The identity gate does nothing."""
        return []

    @decompose_op.register(Hadamard)
    def _(self, gate: Hadamard):
        r"""Decomposes the Hadamard as :math:`H = R_{y}(\pi/2) Z`."""
        return [Z(qubit=gate.qubit), Ry(qubit=gate.qubit, theta=np.pi / 2)]

    @decompose_op.register(S)
    def _(self, gate: S):
        r"""Decompose the :math:`S` gate as a :math:`R_{z}(\pi/2)` gate (up to global
        phase).
        """
        return [Rz(qubit=gate.qubit, theta=np.pi / 2)]

    @decompose_op.register(Sdg)
    def _(self, gate: Sdg):
        r"""Decompose the :math:`S^{\dagger}` gate as a :math:`R_{z}(-\pi/2)` gate (up to
        global phase).
        """
        return [Rz(qubit=gate.qubit, theta=-np.pi / 2)]

    @decompose_op.register(T)
    def _(self, gate: T):
        r"""Decompose the :math:`T` gate as a :math:`R_{z}(\pi/4)` gate (up to global
        phase).
        """
        return [Rz(qubit=gate.qubit, theta=np.pi / 4)]

    @decompose_op.register(Tdg)
    def _(self, gate: Tdg):
        r"""Decompose the :math:`T^{\dagger}` gate as a :math:`R_{z}(-\pi/4)` gate (up to
        global phase).
        """
        return [Rz(qubit=gate.qubit, theta=-np.pi / 4)]

    @decompose_op.register(ECR)
    def _(self, gate: ECR):
        r"""Decomposes an ECR using :math:`ZX^{0, 1}(-\pi/4) R_{x}^{0} ZX^{0, 1}(\pi/4)`."""
        return [
            ZX_pi_4(qubit1=gate.qubit1, qubit2=gate.qubit2),
            Rx(qubit=gate.qubit1, theta=np.pi),
            Rz(qubit=gate.qubit1, theta=np.pi),
            Rz(qubit=gate.qubit2, theta=np.pi),
            ZX_pi_4(qubit1=gate.qubit1, qubit2=gate.qubit2),
            Rz(qubit=gate.qubit1, theta=np.pi),
            Rz(qubit=gate.qubit2, theta=np.pi),
        ]

    @decompose_op.register(CNOT)
    def _(self, gate: CNOT):
        r"""Decompose a CNOT into an ECR gate as
        :math:`R_{x}^{1}(-\pi/2) R_{Z}^{0}(-\pi/2) R_{x}^{0}(\pi) ECR^{0, 1}`.
        """
        return [
            ECR(qubit1=gate.control, qubit2=gate.target),
            Rx(qubit=gate.control, theta=np.pi),
            Rz(qubit=gate.control, theta=-np.pi / 2),
            Rx(qubit=gate.target, theta=-np.pi / 2),
        ]

    @decompose_op.register(SWAP)
    def _(self, gate: SWAP):
        """Decompose a SWAP gate as three CNOTS."""
        return [
            CNOT(qubit1=gate.qubit1, qubit2=gate.qubit2),
            CNOT(qubit1=gate.qubit2, qubit2=gate.qubit1),
            CNOT(qubit1=gate.qubit1, qubit2=gate.qubit2),
        ]

    @decompose_op.register(CRz)
    def _(self, gate: CRz):
        """Decomposes a controlled-Rz gate into two CNOTs and two Rz gates."""
        return [
            Rz(qubit=gate.target, theta=gate.theta / 2),
            CNOT(qubit1=gate.control, qubit2=gate.target),
            Rz(qubit=gate.target, theta=-gate.theta / 2),
            CNOT(qubit1=gate.control, qubit2=gate.target),
        ]

    @decompose_op.register(CPhase)
    def _(self, gate: CPhase):
        """Decomposes a controlled-Phase gate into a controlled-Rz gate and an Rz gate."""
        return [
            Rz(qubit=gate.control, theta=gate.theta / 2),
            CRz(qubit1=gate.control, qubit2=gate.target, theta=gate.theta),
        ]

    @decompose_op.register(CRx)
    def _(self, gate: CRx):
        """Decomposes a controlled-Rx gate via controlled-Rz gates and Hadamards"""
        return [
            Hadamard(qubit=gate.target),
            CRz(qubit1=gate.control, qubit2=gate.target, theta=gate.theta),
            Hadamard(qubit=gate.target),
        ]

    @decompose_op.register(CRy)
    def _(self, gate: CRy):
        """Decomposes a controlled-Ry gate via cnots and Ry gates."""
        return [
            Ry(qubit=gate.target, theta=gate.theta / 2),
            CNOT(qubit1=gate.control, qubit2=gate.target),
            Ry(qubit=gate.target, theta=-gate.theta / 2),
            CNOT(qubit1=gate.control, qubit2=gate.target),
        ]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.pipelines.legacy.base import LegacyExecutePipeline as LegacyQbloxExecutePipeline
from qat.pipelines.updateable import PipelineConfig
from qat.purr.compiler.hardware_models import QuantumHardwareModel

from .compile import LegacyQbloxCompilePipeline
from .full import LegacyQbloxPipeline


def get_pipeline(
    model: QuantumHardwareModel, name: str = "legacy_qblox"
) -> LegacyQbloxPipeline:
    """A factory for creating pipelines that performs the responsibilities of the legacy
    runtime :class:`EchoEngine <qat.purr.compiler.runtime.QuantumRuntime>`

    :param model: The Qblox hardware model.
    :param name: The name of the pipeline, defaults to "legacy_qblox"
    :return: The complete pipeline, including the runtime and engine.
    """
    return LegacyQbloxPipeline(config=PipelineConfig(name=name), model=model).pipeline


__all__ = [
    "LegacyQbloxPipeline",
    "LegacyQbloxCompilePipeline",
    "LegacyQbloxExecutePipeline",
    "PipelineConfig",
    "get_pipeline",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from abc import ABC, abstractmethod

from qat.model.hardware_model import PhysicalHardwareModel
from qat.purr.compiler.execution import QuantumExecutionEngine
from qat.purr.compiler.hardware_models import QuantumHardwareModel


class RequiresHardwareModelMixin(ABC):
    """Specifies a hardware model requirement for a :class:`NativeEngine`. To be used as a
    mixin for engines that require a hardware model to be set.

    Engines might require a hardware model to be set to execute. This might be because the
    hardware model contains information about the control hardware of the target device
    (for the case of legacy hardware models), or because the engine is a simulator that
    makes use of the calibration data.
    """

    @property
    def model(self) -> PhysicalHardwareModel | QuantumHardwareModel:
        return self._model

    @model.setter
    def model(self, model: PhysicalHardwareModel | QuantumHardwareModel):
        """Uses indirection for the model setter, so the child class can implement its
        own logic for updating the model."""

        if self._model != model:
            self._update_model(model)

    @abstractmethod
    def _update_model(self, model: PhysicalHardwareModel | QuantumHardwareModel):
        """Method to update the model in the engine. This is to be implemented by the child
        class, as changing the model might have different implications."""
        ...


def requires_hardware_model(engine) -> bool:
    """Checks if an engine requires a hardware model. This is used to allow backwards
    compatibility with purr engines, but should be removed in the future."""

    # TODO: Yeet this function in the bin for v4 (COMPILER-662)

    if isinstance(engine, RequiresHardwareModelMixin):
        return True
    if isinstance(engine, QuantumExecutionEngine) and engine.model is not None:
        return True
    return False



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from compiler_config.config import CompilerConfig

from qat.core.metrics_base import MetricsManager
from qat.core.result_base import ResultManager
from qat.frontend.base import BaseFrontend
from qat.purr.compiler.builders import InstructionBuilder


class FallthroughFrontend(BaseFrontend):
    """
    A frontend that passes through an input :class:`InstructionBuilder` and does not modify it.
    Used in situations where a frontend is not required, but is used to make a pipeline complete.
    """

    # TODO: add support for OPTIONAL type checking. We do not always want anything to
    # "fall through" here, we might only want particular types, e.g., Qat IR (COMPILER-333)

    def emit(
        self,
        src: InstructionBuilder,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ) -> InstructionBuilder:
        return src

    def check_and_return_source(self, src):
        """All source files are valid for the :class:`FallthroughFrontend`."""
        return src



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from inspect import signature

from qat.core.config.descriptions import PipelineFactoryDescription
from qat.engines import NativeEngine
from qat.model.target_data import AbstractTargetData
from qat.pipelines.base import BasePipeline
from qat.pipelines.updateable import Model, UpdateablePipeline


class PipelineFactory(UpdateablePipeline):
    """An updateable pipeline that allows a pipeline factory (function) to be called.
    This will use the factory as a proxy for the _build_pipeline method, allowing the
    pipeline to be refreshed with a new hardware model and/or config."""

    @staticmethod
    def _build_pipeline(
        config: PipelineFactoryDescription,
        model: Model,
        target_data: AbstractTargetData | None = None,
        engine: NativeEngine | None = None,
    ) -> BasePipeline:
        """Wraps the pipeline factory function defined in the config, passing the
        model and target data if required."""

        factory = config.pipeline
        kwargs = {}
        if PipelineFactory._has_argument(factory, "engine"):
            kwargs["engine"] = engine
        if PipelineFactory._has_argument(factory, "target_data"):
            kwargs["target_data"] = target_data
        kwargs.update(config.config)
        return factory(model=model, **kwargs)

    @staticmethod
    def _has_argument(factory: callable, arg: str):
        """Checks if the factory accepts the requested argument."""
        return arg in signature(factory).parameters



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from functools import singledispatchmethod

from qat.ir.gates.base import QubitInstruction


class DecompositionBase:
    """Decomposition objects implement the rules for decomposing gates into native gates.

    The object should be equipped with a :meth:`decompose_op` method that is implemented
    using single dispatch. This allows us to register a decomposition for each relevant gate
    / operation type. Note that decompositions should be implemented carefully, and form a
    directed acyclic graph (DAG) that terminates at :class:`NativeGate` nodes.

    Assumming all relevant nodes are implemented, an operation can be decomposed to a series
    of native gates / operations using the :meth:`decompose` method, which will recursively
    call :meth:`decompose_op`.

    As a rule of thumb, try not to use logic on gate angles to define the decomposition.
    They should be kept general. If there are instances where a choice of angle(s) can be
    decomposed more effeciently, a transformation pass should be used to substitute the
    gate. This is done to allow predictable behaviour, and has parameterised circuits in
    mind.
    """

    end_nodes = tuple()

    def __init__(
        self,
        end_nodes: list[QubitInstruction] | None = None,
        extend_end_nodes: list[QubitInstruction] | None = None,
    ):
        """
        :param end_nodes: Optionally specify the nodes to end the decomposition recursion
            at. Defaults to :code:`decompositions.end_nodes`.
        :param extend_end_nodes: By default the decomposition will recurse until all gates
            are a :class:`NativeGate`. You can optionally add additional (non
            :class:`NativeGate`) gates to terminate at. This argument should only be
            specified if :code:`end_nodes=None`.
        """

        if end_nodes:
            self.end_nodes = tuple(set(end_nodes))
        elif extend_end_nodes:
            end_nodes = set(self.end_nodes)
            end_nodes.update(extend_end_nodes)
            self.end_nodes = tuple(end_nodes)

    def decompose(self, gate: QubitInstruction, *args):
        """Recursively decomposes an instruction into a product of native gates.

        :param gate: Gate to decompose.
        """
        if isinstance(gate, self.end_nodes):
            return [gate]

        gate_list = []
        for g in self.decompose_op(gate, *args):
            gate_list.extend(self.decompose(g, *args))

        return gate_list

    @singledispatchmethod
    def decompose_op(self, gate: QubitInstruction, *args):
        """Implements the definition of a decomposition of a qubit operation.

        The definition does not have to be in terms of end nodes, but decompositions must
        form a DAG. For example,

        .. code-block:: python

            CNOT -> {ECR, X, Z} -> {ZX_pi_4, Z_phase, X_pi_2}
            U -> {x_pi_2, Z_phase}
        """
        raise NotImplementedError(
            f"Decomposition for gate {gate.__class__.__name__} not implemented."
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.backend.waveform_v1.executable import (
    PositionalAcquireData as PositionalAcquireData,
    WaveformV1ChannelData as WaveformV1ChannelData,
    WaveformV1Program as WaveformV1Program,
)
from qat.backend.waveform_v1.purr.codegen import WaveformV1Backend as WaveformV1Backend



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from .base import CustomMiddleend as CustomMiddleend
from .default import PydDefaultMiddleend as PydDefaultMiddleend
from .fallthrough import FallthroughMiddleend as FallthroughMiddleend
from .purr.default import DefaultMiddleend as DefaultMiddleend



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.core.pass_base import PassManager
from qat.model.target_data import TargetData
from qat.pipelines.pipeline import ExecutePipeline
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.backends.qiskit_simulator import QiskitHardwareModel
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.utils.logger import get_default_logger
from qat.runtime.legacy import LegacyRuntime
from qat.runtime.passes.purr.transform import (
    QiskitErrorMitigation,
    QiskitSimplifyResults,
    QiskitStripMetadata,
)

log = get_default_logger()


def results_pipeline() -> PassManager:
    return (
        PassManager()
        | QiskitStripMetadata()
        | QiskitErrorMitigation()
        | QiskitSimplifyResults()
    )


class LegacyQiskitExecutePipeline(UpdateablePipeline):
    """A pipeline that executes programs using the :class:`QiskitEngine` and the
    :class:`LegacyRuntime`.

    Implements a custom pipeline to make instructions suitable for the legacy Qiskit engine,
    and has a custom post-processing pipeline.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: QuantumHardwareModel,
        target_data: TargetData | None = None,
        engine: None = None,
    ) -> ExecutePipeline:
        if not isinstance(model, QiskitHardwareModel):
            raise TypeError("Model must be an instance of QiskitHardwareModel")

        if engine is not None:
            log.warning(
                "An engine was provided to the LegacyQiskitExecutePipeline, but it will be "
                "ignored. The legacy Qiskit engine is used directly."
            )

        return ExecutePipeline(
            name=config.name,
            model=model,
            target_data=target_data if target_data is not None else TargetData.default(),
            runtime=LegacyRuntime(
                engine=model.create_engine(),
                results_pipeline=results_pipeline(),
            ),
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from collections import defaultdict
from dataclasses import dataclass, field

from qat.ir.instructions import Assign, Instruction, ResultsProcessing, Return
from qat.ir.measure import Acquire, PostProcessing
from qat.ir.pulse_channel import PulseChannel


@dataclass
class PartitionedIR:
    """Contains the IR that has been processed and reduced to a program per pulse channel.
    Also splits out any post-processing instructions to be used by the executor.

    Programs are typically sent to hardware as a package per physical / logical channel,
    and need to be partitioned out. They also need to be free of instructions that are not
    used at runtime. This class does this.

    This will eventually be replaced by some "partitoned module" approach, which might have
    a similar schema, or might just be many modules.
    """

    target_map: dict[str, list[Instruction]] = field(
        default_factory=lambda: defaultdict(list)
    )
    pulse_channels: dict[str, PulseChannel] = field(default_factory=dict)
    shots: int | None = field(default=None)
    compiled_shots: int | None = field(default=None)
    returns: list[Return] = field(default_factory=list)
    assigns: list[Assign] = field(default_factory=list)
    acquire_map: dict[PulseChannel | str, list[Acquire]] = field(
        default_factory=lambda: defaultdict(list)
    )
    pp_map: dict[str, list[PostProcessing]] = field(
        default_factory=lambda: defaultdict(list)
    )
    rp_map: dict[str, ResultsProcessing] = field(default_factory=dict)

    def get_pulse_channel(self, id: str) -> PulseChannel | None:
        """Get a pulse channel by its ID.

        Maintains API with the QuantumInstructionBuilder."""
        return self.pulse_channels.get(id, None)
